"""Functions for plotting qq-plot.

Copytright (c) Shujia Huang
Date: 2021-02-21

"""
import numpy as np
from scipy.stats import norm
from matplotlib.pyplot import subplots

from ..utils import is_numeric


def ppoints(n, a=0.5):
    """ 
    numpy analogue `R`'s `ppoints` function.

    ppoints() is used in qqplot and qqnorm to generate the set of 
    probabilities at which to evaluate the inverse distribution.

    see details at 
        http://stat.ethz.ch/R-manual/R-patched/library/stats/html/ppoints.html 

    Parameters
    ----------
    n : a number or array type
        If n is an array or list, ``n`` will re-assign to ``n=len(n)``
    
    a : float, optional, default: 0.5
        The offset fraction to be used; typically in (0, 1)
        Recommend to set: a = 3.0/8.0 if n <= 10 else 0.5 


    Returns
    -------
        A list of value calculates by this formular:
        (1:n - a)/(n + (1-2a)
        Typically it's an array with elements in (0, 1)


    Notes
    -----
    * Becker, R. A., Chambers, J. M. and Wilks, A. R. (1988) The New S Language. 
      Wadsworth & Brooks/Cole.
    * Blom, G. (1958) Statistical Estimates and Transformed Beta Variables. Wiley

    """

    if a < 0 or a > 1:
        msg = "`a` could just be any float value in (0, 1)."
        raise ValueError(msg)

    try:
        n = len(n)
    except TypeError:
        pass

    # ``n`` is a float value
    return (np.arange(n, dtype=float) + 1 - a) / (n + 1 - 2 * a)


def qqplot(data, other=None, logp=True, ax=None, marker="o", color=None, alpha=0.8, title=None,
           xlabel=None, ylabel=None, ablinecolor="r", **kwargs):
    """Creat Q-Q plot.
    **CAUSION: The x-axis(expected) is created from uniform distribution.**

    Parameters
    ----------
    data : list, 1d-array-like, or Series
        Data to be plotted

    other : list, 1d-array-like, Series or None, optional
        If provided, the sample quantiles of the `data` array-like object are 
        plotted against the sample quantiles of the `other` array-like object. 
        If not provided (default), the theoretical quantiles are used.

    logp : bool, default is 'True', optional
        If true, -log10 of the y_value(always be the p-value) is plotted.
        It isn't very useful to plot raw p-values in GWAS QQ plot.

    ax : matplotlib axis, optional
        Axis to plot on, otherwise uses current axis.

    marker : matplotlib markers for scatter plot, default is "o", optional

    color : matplotlib color, optional
        The dots color in the plot

    alpha : float scalar, default is 0.8, optional
        The alpha blending value, between 0(transparent) and 1(opaque)

    title : string, or None, optional
        Set the title of the current plot.

    xlabel : string, or None, optional
        Set the x axis label of the current axis.
        CAUSION: The x axis will always be the expected value.

    ylabel : string, or None, optional
        Set the y axis label of the current axis.
        CAUSION: The y axis will always be the observed value.

    ablinecolor : matplotlib color, default is 'r' (red), optional
        Color for the abline in plot. if set ``ablinecolor=None`` 
        means do not plot the abline.

    kwargs : key, value pairings, optional
        Other keyword arguments are passed to ``plt.scatter()``
        (in matplotlib.pyplot).


    Returns
    -------
    ax : matplotlib Axes
        Axes object with the plot.

    Notes
    -----
    1. The X axis will always be the expected values and Y axis always be
       observed values in the plot.
    2. This plot function is not just suit for GWAS QQ plot, it could
       also be used for creating QQ plot for other data, which format 
       are list-like ::
        [value1, value2, ...] (all the values should between 0 and 1)

    See Also
    --------
    qqnorm : Q-Q plot against the normal distribution.

    Examples
    --------

    Plot a basic QQ plot:

    .. plot::
        :context: close-figs

        >>> import pandas as pd
        >>> from geneview import qqplot
        >>> from geneview.utils import load_dataset
        >>> df = load_dataset("gwas")
        >>> ax = qqplot(data=df["P"])

    Plot a better QQ plot with title and label texts.

    .. plot::
        :context: close-figs

        >>> from matplotlib.pyplot import subplots
        >>> fig, ax = subplots(figsize=(6, 6), facecolor="w", edgecolor="k")  # define a plot
        >>> ax = qqplot(data=df["P"],
        ...             marker="o",
        ...             title="Test",
        ...             xlabel=r"Expected $-log_{10}{(P)}$",
        ...             ylabel=r"Observed $-log_{10}{(P)}$",
        ...             ax=ax)

    We could even create a QQ plot base on two different dataset:

    .. plot::
        :context: close-figs

        >>> import numpy as np
        >>> data1 = np.random.normal(size = 100)
        >>> data2 = np.random.normal(5.0, 1.0, size = 100)
        >>> ax = qqplot(data=data1, other=data2, logp=False, xlabel="Expected", ylabel="Observe")
    """
    if not all(map(is_numeric, data)):
        msg = 'Input must all be numeric in `data`.'
        raise ValueError(msg)

    if other is not None and not all(map(is_numeric, other)):
        msg = 'Input must all be numeric in `other`.'
        raise ValueError(msg)

    if other is not None and len(other) != len(data):
        msg = 'Input `data` and `other` must all be the same size.'
        raise ValueError(msg)

    if xlabel is None:
        xlabel = r"$Expected(-log_{10}{(P)})$" if other is None else r"$-log_{10}{(Value)} of 2nd Sample$"
    if ylabel is None:
        ylabel = r"$Observed(-log_{10}{(P)})$" if other is None else r"(-log_{10}{(Value)}) of 1st Sample$"

    data = np.array(data, dtype=float)

    # create observed and expected
    e = ppoints(len(data)) if other is None else sorted(other)

    if logp:
        o = -np.log10(sorted(data))
        e = -np.log10(e)
    else:
        o = np.array(sorted(data))
        e = np.array(e)

    if "marker" not in kwargs:
        kwargs["marker"] = marker
    ax = _do_plot(e, o, ax=ax, color=color, ablinecolor=ablinecolor, alpha=alpha, **kwargs)

    lambda_value = round(np.median(norm.ppf(data) ** 2 / 0.454), 3)
    if title:
        title += r"$(\lambda = %s)$" % lambda_value
    else:
        title = r"$\lambda = %s$" % lambda_value

    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    return ax


def qqnorm(data, ax=None, xlabel="Expected", ylabel="Observed", color=None, ablinecolor="r",
           alpha=0.8, **kwargs):
    """Creat Q-Q plot against the normal distribution values.
    *CAUSION: The x-axis(expected) is created from normal distribution.*

    Parameters
    ----------
    data : array-like. ``Series`` of ``pandas`` or 1d array-like
        The data will be normalization be this formula:
        (data - data.mean()) / data.std() before to be plotted.

    ax : matplotlib axis, optional
        Axis to plot on, otherwise uses current axis.

    xlabel: string, or 'Expected', optional
        Set the x axis label of the current axis.
        CAUSION: The x axis will always be the expected value.

    ylabel: string, or 'Observed', optional
        Set the y axis label of the current axis.
        CAUSION: The y axis will always be the observed value.

    color : matplotlib color, optional
        The dots color in the plot

    ablinecolor: matplotlib color, default is 'r' (red), optional
        Color for the abline in plot. if set ``ablinecolor=None`` 
        means do not plot the abline.

    alpha : float scalar, default is 0.8, optional
        The alpha blending value, between 0(transparent) and 1(opaque)

    kwargs : key, value pairings
        Other keyword arguments are passed to ``plt.scatter()``
        (in matplotlib.pyplot).


    Returns
    -------
    ax : matplotlib Axes
        Axes object with the plot.


    Notes
    -----
    1. The X axis will always be the expected values and Y axis always be
       observed values in the plot.


    See Also
    --------
    qqplot : Q-Q plot against other values or uniform distribution(default)


    Examples
    --------

    Plot a basic QQ norm plot:

    .. plot::
        :context: close-figs

        >>> import numpy as np
        >>> data = np.random.normal(size = 100)
        >>> ax = qqnorm(data=data)

    Plot a QQ norm plot with GOYA_preview data:

    .. plot::
        :context: close-figs

        >>> import pandas as pd
        >>> from geneview.utils import load_dataset
        >>> df = load_dataset("gwas")
        >>> ax = qqnorm(data=df["P"], xlabel="Expected value", ylabel="Observed value")
    """
    if not all(map(is_numeric, data)):
        msg = 'Input must all be numeric in `data`.'
        raise ValueError(msg)

    # Normalization the data to be in (mu=0.0, std=1.0) normal distribution
    obs = np.array(data, dtype=float)
    obs = (obs - obs.mean()) / obs.std()
    obs.sort()

    # create expected
    e = norm.ppf(ppoints(len(obs)))

    ax = _do_plot(e, obs, ax=ax, color=color, ablinecolor=ablinecolor,
                  alpha=alpha, **kwargs)

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    return ax


def _do_plot(x, y, ax=None, color=None, ablinecolor="r", alpha=0.8, **kwargs):
    """
    Boiler plate plotting function for the `qqplot` and `qqnorm`

    Parameters
    ----------
    x, y : array-like
        Data to be plotted

    ax : matplotlib axis, optional
        Axis to plot on, otherwise uses current axis.

    color : matplotlib color, optional
        The dots color in the plot
        
    ablinecolor: matplotlib color, default is 'r' (red), optional
        Color for the abline in plot. if set ``ablinecolor=None`` 
        means do not plot the abline.

    alpha : float scalar, default is 0.8, optional
        The alpha blending value, between 0(transparent) and 1(opaque)

    kwargs : key, value pairings
        Other keyword arguments are passed to ``plt.scatter()``
        (in matplotlib.pyplot).

    Returns
    -------
    ax : matplotlib Axes
        Axes object with the plot.
    """
    # Draw the plot and return the Axes
    if ax is None:
        # ax = plt.gca()
        _, ax = subplots(figsize=(5, 5), facecolor="w", edgecolor="k")

    # Get the color from the current color cycle
    if color is None:
        line, = ax.plot(0, x.mean())
        color = line.get_color()
        line.remove()

    # x is for expected; y is for observed value
    ax.scatter(x, y, c=color, alpha=alpha, edgecolors='none', **kwargs)
    ax.set_xlim(xmin=x.min(), xmax=1.05 * x.max())
    ax.set_ylim(ymin=y.min())

    if ablinecolor:
        # plot the y=x line by expected: uniform distribution data
        ax.plot([x.min(), ax.get_xlim()[1]], [x.min(), ax.get_xlim()[1]],
                color=ablinecolor, linestyle="-")

    return ax



import colorsys
from distutils.version import LooseVersion

import matplotlib as mpl
import matplotlib.colors as mplcol

mpl_ge_150 = LooseVersion(mpl.__version__) >= "1.5.0"

def desaturate(color, prop):
    """Decrease the saturation channel of a color by some percent.

    Parameters
    ----------
    color : matplotlib color
        hex, rgb-tuple, or html color name
    prop : float
        saturation channel of color will be multiplied by this value

    Returns
    -------
    new_color : rgb tuple
        desaturated color code in RGB tuple representation

    """
    # Check inputs
    if not 0 <= prop <= 1:
        raise ValueError("prop must be between 0 and 1")

    # Get rgb tuple rep
    rgb = mplcol.colorConverter.to_rgb(color)

    # Convert to hls
    h, l, s = colorsys.rgb_to_hls(*rgb)

    # Desaturate the saturation channel
    s *= prop

    # Convert back to rgb
    new_color = colorsys.hls_to_rgb(h, l, s)

    return new_color


def set_hls_values(color, h=None, l=None, s=None):
    """Independently manipulate the h, l, or s channels of a color.

    Parameters
    ----------
    color : matplotlib color
        hex, rgb-tuple, or html color name
    h, l, s : floats between 0 and 1, or None
        new values for each channel in hls space

    Returns
    -------
    new_color : rgb tuple
        new color code in RGB tuple representation

    """
    # Get rgb tuple representation
    rgb = mplcol.colorConverter.to_rgb(color)
    vals = list(colorsys.rgb_to_hls(*rgb))
    for i, val in enumerate([h, l, s]):
        if val is not None:
            vals[i] = val

    rgb = colorsys.hls_to_rgb(*vals)
    return rgb


def get_color_cycle():
    if mpl_ge_150:
        cyl = mpl.rcParams['axes.prop_cycle']
        # matplotlib 1.5 verifies that axes.prop_cycle *is* a cycler
        # but no garuantee that there's a `color` key.
        # so users could have a custom rcParmas w/ no color...
        try:
            return [x['color'] for x in cyl]
        except KeyError:
            pass  # just return axes.color style below
    return mpl.rcParams['axes.color_cycle']



"""
Plotting functions for create karyotype plots.

Copyright (c) Shujia Huang
Date: 2016-02-19
"""
import numpy as np
from pandas import DataFrame
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle

from ..palette import circos  # ``circos`` is a color dict


def karyoplot(data, ax=None, width=0.5, CHR=None, alpha=0.8, color4none="#34728B", **kwargs):
    """ Create karyotype plot.

    Parameters
    ----------
    data : string or array
        A karyotype information list or input file path, even more could be 
        any kind of URL link. e.g. AWS S3 link.

    ax : matplotlib axis, optional
        Axis to plot on, otherwise uses current axis.

    width : float, optional, default: 0.5
        Chromosom"s width in the plot

    CHR : string, optional, defualt: None
        Choice the specific chromosome to plot.

    alpha : scalar, optional, default: 0.8   
        The alpha blending value, between 0(transparent) and 1(opaque)

    color4none : matplotlib color, optional, default: "#34728B"(deep gray blue)
        The color for undefine band color of karyotype in the plot.

    kwargs : key, value pairings
        Other keyword arguments are passed to ``Rectangle`` in matplotlib.patches

    Examples
    --------

    A basic karyotype plot get the input karyotype information from URL:

    .. plot::
        :context: close-figs

        >>> import matplotlib.pyplot as plt
        >>> from geneview.utils import load_dataset
        >>> from geneview import karyoplot
        >>> fig, ax = plt.subplots(figsize=(20, 5))
        >>> k_fn = load_dataset("karyotype_human_hg19.txt")
        >>> _ = karyoplot(k_fn, ax=ax)

    """
    # Draw the plot and return the Axes 
    if ax is None:
        ax = plt.gca()

    if isinstance(data, str):
        # suppose to be a path to the input file or a url to the file
        data = pd.read_table(data, header=0,
                             names=["chrom", "start", "end", "name", "gie_stain"])
    elif isinstance(data, DataFrame):
        # reset the columns
        data = DataFrame(data.values, columns=["chrom", "start", "end", "name", "gie_stain"])
    else:
        # convert to DataFrame of pandas
        data = DataFrame(data, columns=["chrom", "start", "end", "name", "gie_stain"])

    yaxis = []
    for i, (chrom, kc_df) in enumerate(sorted(data.groupby("chrom"), key=lambda x: x[0])):

        if CHR is not None and chrom != CHR:
            continue

        yaxis.append(chrom)
        for _, r in kc_df.iterrows():
            band_color = circos[r.gie_stain] if r.gie_stain in circos else color4none
            band_rec = Rectangle((r.start, i), r.end - r.start, width,
                                 color=band_color, **kwargs)
            ax.add_patch(band_rec)

    xmax = data["end"].max() * 1.1
    xticks = np.arange(0, xmax, xmax / 10.)
    ax.set_xticks(xticks)
    ax.set_xticklabels(["{0}M".format(int(i / 10 ** 6)) for i in xticks])
    ax.set_xlim(0, xmax)

    ax.set_yticks([i + width / 2 for i in range(len(yaxis))])
    ax.set_yticklabels(yaxis)
    ax.set_ylim(0, len(yaxis))

    return ax



"""
Small plotting-related utility functions.
=========================================

Copy from: the seaborn repo's utils.py on github, Then we make some modifies 
for geneview.

"""
import matplotlib.pyplot as plt


class General(object):
    def __init__(self):
        pass

    @staticmethod
    def get_figure(is_show, fig_name=None, dpi="figure"):
        if fig_name:
            plt.savefig(fig_name, bbox_inches="tight", dpi=dpi)

        if is_show:
            plt.show()

        plt.clf()
        plt.close()

        return



from inspect import signature, Parameter
from functools import wraps
import warnings


# This function was adapted from scikit-learn
# github.com/scikit-learn/scikit-learn/blob/master/sklearn/utils/validation.py
def deprecate_positional_args(f):
    """Decorator for methods that issues warnings for positional arguments.
    Using the keyword-only argument syntax in pep 3102, arguments after the
    * will issue a warning when passed as a positional argument.
    Parameters
    ----------
    f : function
        function to check arguments on
    """
    sig = signature(f)
    kwonly_args = []
    all_args = []

    for name, param in sig.parameters.items():
        if param.kind == Parameter.POSITIONAL_OR_KEYWORD:
            all_args.append(name)
        elif param.kind == Parameter.KEYWORD_ONLY:
            kwonly_args.append(name)

    @wraps(f)
    def inner_f(*args, **kwargs):
        extra_args = len(args) - len(all_args)
        if extra_args > 0:
            plural = "s" if extra_args > 1 else ""
            article = "" if plural else "a "
            warnings.warn(
                "Pass the following variable{} as {}keyword arg{}: {}. "
                "From version 0.12, the only valid positional argument "
                "will be `data`, and passing other arguments without an "
                "explicit keyword will result in an error or misinterpretation."
                .format(plural, article, plural,
                        ", ".join(kwonly_args[:extra_args])),
                FutureWarning
            )
        kwargs.update({k: arg for k, arg in zip(sig.parameters, args)})
        return f(**kwargs)
    return inner_f


def share_init_params_with_map(cls):
    """Make cls.map a classmethod with same signature as cls.__init__."""
    map_sig = signature(cls.map)
    init_sig = signature(cls.__init__)

    new = [v for k, v in init_sig.parameters.items() if k != "self"]
    new.insert(0, map_sig.parameters["cls"])
    cls.map.__signature__ = map_sig.replace(parameters=new)
    cls.map.__doc__ = cls.__init__.__doc__

    cls.map = classmethod(cls.map)

    return cls



"""
This module contains miscellaneous functions for ``geneview``.
"""
import operator
import pandas as pd
import numpy as np
from scipy import stats


def is_numeric(s):
    """
    It's a useful function for checking if a data is a numeric.

    This function could identify any kinds of numeric: e.g. '0.1', '10', '-2.',
    2.5, etc

    Parameters
    ----------
    s : int, float or string.
        The input could be any kind of single value except the scalable
        type of python like 'list()', 'dict()', 'set()'

    Returns
    -------
        A boolean. Ture if `s` is numeric else False

    Notes
    -----
        http://stackoverflow.com/questions/354038/how-do-i-check-if-a-string-is-a-number-float-in-python
    """
    try:
        float(s)
        return True
    except ValueError:
        return False


def is_integer(s):
    """
    This function could identify any kinds of integer

    Parameters
    ----------
        s : int, float or string.
            The input could be any kind of single value except the scalable
            type of python like 'list()', 'dict()', 'set()'

    Returns
    -------
        A boolean. Ture if `s` is integer value else False 
    """
    if is_numeric(s):
        return True if '.' not in s else False
    else:
        return False


def iqr(a):
    """Calculate the IQR for an array of numbers."""
    a = np.asarray(a)
    q1 = stats.scoreatpercentile(a, 25)
    q3 = stats.scoreatpercentile(a, 75)
    return q3 - q1


def _kde_support(data, bw, gridsize, cut, clip):
    """Establish support for a kernel density estimate."""
    support_min = max(data.min() - bw * cut, clip[0])
    support_max = min(data.max() + bw * cut, clip[1])
    return np.linspace(support_min, support_max, gridsize)


def categorical_order(values, order=None):
    """Return a list of unique data values.

    Determine an ordered list of levels in ``values``.

    Parameters
    ----------
    values : list, array, Categorical, or Series of pandas
        Vector of "categorical" values
    order : list-like, optional
        Desired order of category levels to override the order determined
        from the ``values`` object.

    Returns
    -------
    order : list
        Ordered list of category levels not including null values.

    """
    if order is None:
        if hasattr(values, "categories"):
            order = values.categories
        else:
            try:
                order = values.cat.categories
            except (TypeError, AttributeError):
                try:
                    order = values.unique()
                except AttributeError:
                    order = pd.unique(values)
                try:
                    np.asarray(values).astype(np.float)
                    order = np.sort(order)
                except (ValueError, TypeError):
                    order = order
        order = filter(pd.notnull, order)
    return list(order)


def freedman_diaconis_bins(a):
    """Calculate number of hist bins using Freedman-Diaconis rule."""
    # From http://stats.stackexchange.com/questions/798/
    a = np.asarray(a)
    h = 2 * iqr(a) / (len(a) ** (1 / 3))
    # fall back to sqrt(a) bins if iqr is 0
    if h == 0:
        return int(np.sqrt(a.size))
    else:
        return int(np.ceil((a.max() - a.min()) / h))



"""
Author: Shujia Huag
Date: 2021-05-11 11:49:59
"""
import pytest

from ..utils import deprecate_positional_args


# This test was adapted from scikit-learn
# github.com/scikit-learn/scikit-learn/blob/master/sklearn/utils/tests/test_validation.py
def test_deprecate_positional_args_warns_for_function():
    @deprecate_positional_args
    def f1(a, b, *, c=1, d=1):
        return a, b, c, d

    with pytest.warns(
            FutureWarning,
            match=r"Pass the following variable as a keyword arg: c\."
    ):
        assert f1(1, 2, 3) == (1, 2, 3, 1)

    with pytest.warns(
            FutureWarning,
            match=r"Pass the following variables as keyword args: c, d\."
    ):
        assert f1(1, 2, 3, 4) == (1, 2, 3, 4)

    @deprecate_positional_args
    def f2(a=1, *, b=1, c=1, d=1):
        return a, b, c, d

    with pytest.warns(
            FutureWarning,
            match=r"Pass the following variable as a keyword arg: b\.",
    ):
        assert f2(1, 2) == (1, 2, 1, 1)

    # The * is placed before a keyword only argument without a default value
    @deprecate_positional_args
    def f3(a, *, b, c=1, d=1):
        return a, b, c, d

    with pytest.warns(
            FutureWarning,
            match=r"Pass the following variable as a keyword arg: b\.",
    ):
        assert f3(1, 2) == (1, 2, 1, 1)


def test_deprecate_positional_args_warns_for_class():
    class A1:
        @deprecate_positional_args
        def __init__(self, a, b, *, c=1, d=1):
            self.a = a, b, c, d

    with pytest.warns(
            FutureWarning,
            match=r"Pass the following variable as a keyword arg: c\."
    ):
        assert A1(1, 2, 3).a == (1, 2, 3, 1)

    with pytest.warns(
            FutureWarning,
            match=r"Pass the following variables as keyword args: c, d\."
    ):
        assert A1(1, 2, 3, 4).a == (1, 2, 3, 4)

    class A2:
        @deprecate_positional_args
        def __init__(self, a=1, b=1, *, c=1, d=1):
            self.a = a, b, c, d

    with pytest.warns(
            FutureWarning,
            match=r"Pass the following variable as a keyword arg: c\.",
    ):
        assert A2(1, 2, 3).a == (1, 2, 3, 1)

    with pytest.warns(
            FutureWarning,
            match=r"Pass the following variables as keyword args: c, d\.",
    ):
        assert A2(1, 2, 3, 4).a == (1, 2, 3, 4)



"""Plotting functions for manhattan plot.

Copyright (c) Shujia Huang
Date: 2021-02-21

This model is based on brentp's script on github:
https://github.com/brentp/bio-playground/blob/master/plots/manhattan-plot.py

Thanks for Brentp's contributions

"""
from itertools import cycle
from pandas import DataFrame
import numpy as np

from matplotlib.pyplot import subplots
from ..utils import adjust_text


# learn something from "https://github.com/reneshbedre/bioinfokit/blob/38fb4966827337f00421119a69259b92bb67a7d0/bioinfokit/visuz.py"
def manhattanplot(data, chrom="#CHROM", pos="POS", pv="P", snp="ID", logp=True, ax=None,
                  marker=".", color="#3B5488,#53BBD5", alpha=0.8,
                  title=None, xlabel="Chromosome", ylabel=r"$-log_{10}{(P)}$",
                  xtick_label_set=None, CHR=None, xticklabel_kws=None,
                  suggestiveline=1e-5, genomewideline=5e-8, sign_line_cols="#D62728,#2CA02C", hline_kws=None,
                  sign_marker_p=None, sign_marker_color="r",
                  is_annotate_topsnp=False, text_kws=None, ld_block_size=50000, **kwargs):
    """Creates a manhattan plot from PLINK assoc output (or any data frame with chromosome, position, and p-value).

    Parameters
    ----------
    data : DataFrame.
        A DataFrame with columns "#CHROM," "POS," "P," and optionally, "SNP."

    chrom : string, default is "#CHROM", optional
        A string denoting the column name for chromosome. Defaults to be PLINK2.x's "#CHROM".
        Said column must be a character.

    pos : string, default is "POS", optional.
        A string denoting the column name for chromosomal position. Default to PLINK2.x's "POS".
        Said column must be numeric.

    pv : string, default is "P", optional.
        A string denoting the column name for chromosomal p-value. Default to PLINK2.x's "P".
        Said column must be float type.

    snp : string, default is "ID", optional.
        A string denoting the column name for the SNP name (rs number) or the column which you want to
        represent the variants. Default to PLINK2.x's "P". Said column should be a character.

    logp : bool, optional
        If TRUE, the -log10 of the p-value is plotted. It isn't very useful
        to plot raw p-values, but plotting the raw value could be useful for
        other genome-wide plots, for example, peak heights, bayes factors, test
        statistics, other "scores," etc. default: True

    ax : matplotlib axis, optional
        Axis to plot on, otherwise uses current axis.

    marker : matplotlib markers for scatter plot, default is "o", optional

    color : matplotlib color, optional, defauft: "#3B5488,#53BBD5"
        Color used for the plot elements. Could hex-code or rgb,
        e.g: '#3B5488,#53BBD5' or 'rb'

    alpha : float scalar, default is 0.8, optional
        The alpha blending value, between 0(transparent) and 1(opaque)

    title : string, or None, optional
        Set the title of the current plot.

    xlabel: string, optional
        Set the x axis label of the current axis.

    ylabel: string, optional
        Set the y axis label of the current axis.

    xtick_label_set : a set. optional
        Set the current x axis ticks of the current axis.

    CHR : string, or None, optional
        Select a specific chromosome to plot. And the x-axis will be the
        position of this chromosome instead of the chromosome id.

        CAUTION: this parameter could not be used with ``xtick_label_set``
                 together.

    xticklabel_kws : key, value pairings, or None, optional
        Other keyword arguments are passed to set xtick labels in
        maplotlib.axis.Axes.set_xticklabels.

    suggestiveline : float or None, default is 1e-5, optional
        Where to draw a suggestive ax.axhline. Set None to be disable.

    genomewideline : float or None, default is 5e-8
        Where to draw a genome-wide significant ax.axhline. Set None to be disable.

    sign_line_cols : matplotlib color, default: "#D62728,#2CA02C", optional.
        Color used for ``suggestiveline`` and ``genomewideline``.
        Could be hex-code or rgb, e.g: "#D62728,#2CA02C" or 'rb'

    hline_kws : key, value pairings, or None, optional
        keyword arguments for plotting ax.axhline(``suggestiveline`` and ``genomewideline``)
        except the "color" key-pair.

    sign_marker_p : float or None, default None, optional.
        A P-value threshold (suggestive to be 1e-6) for marking the significant SNP sites.

    sign_marker_color : matplotlib color, default: "r", optional.
        Define a color code for significant SNP sites.

    is_annotate_topsnp : boolean, default is False, optional.
        Annotate the top SNP or not for the significant locus.

    text_kws: key, value pairings, or None, optional
        keyword arguments for plotting in`` matplotlib.axes.Axes.text(x, y, s, fontdict=None, **kwargs)``

    ld_block_size : integer, default is 50000, optional
        Set the size of LD block which for finding top SNP. And the top SNP's annotation represent the block.

    kwargs : key, value pairings, optional
        Other keyword arguments are passed to ``plt.scatter()`` or
        ``plt.vlines()`` (in matplotlib.pyplot) depending on whether 
        a scatter or line plot is being drawn.


    Returns
    -------
    ax : matplotlib Axes
        Axes object with the manhattanplot.

    Notes
    -----
    1. This plot function is not just suit for GWAS manhattan plot,
       it could also be used for any input data which have [chromo-
       some, position and p-value] dataframe.

    2. The right and top spines of the plot have been set to be
       invisible by hand.

    Examples
    --------

    Plot a basic manhattan plot from PLINK2.x association output and reture the figure:

    .. plot::
        :context: close-figs

        >>> import pandas as pd
        >>> from geneview import manhattanplot
        >>> from geneview.utils import load_dataset
        >>> df = load_dataset("gwas")
        >>> ax = manhattanplot(data=df)

    Plot a basic manhattan plot with horizontal xtick labels:

    .. plot::
        :context: close-figs

        >>> xtick = set(['chr' + i for i in list(map(str, range(1, 10))) + ['11', '13', '15', '18', '21', 'X']])
        >>> ax = manhattanplot(data=df, xlabel="Chromosome", ylabel=r"$-log_{10}{(P)}$", xtick_label_set=xtick)

    Add a horizontal at y position=3 line with linestyle="--" and lingwidth=1.3
    across the axis:

    .. plot::
        :context: close-figs
    
        >>> ax = manhattanplot(data=df, hline_kws={"linestyle": "--", "lw": 1.3}, xlabel="Chromosome",
        ...                    ylabel=r"$-log_{10}{(P)}$",
        ...                    xtick_label_set = xtick)

    Rotate the x-axis ticklabel by setting ``xticklabel_kws``:

    .. plot::
        :context: close-figs

        >>> ax = manhattanplot(data=df,
        ...                    hline_kws={"linestyle": "--", "lw": 1.3},
        ...                    xlabel="Chromosome",
        ...                    ylabel=r"$-log_{10}{(P)}$",
        ...                    xticklabel_kws={"rotation": "vertical"})

    Plot a better one with genome-wide significant mark and annotate the Top SNP and save
    the figure to "output_manhattan_plot.png":

    .. plot::
        :context: close-figs

        >>> from matplotlib.pyplot import subplots
        >>> fig, ax = subplots(figsize=(12, 4), facecolor="w", edgecolor="k")  # define a plot
        >>> ax = manhattanplot(data=df,
        ...                    marker=".",
        ...                    sign_marker_p=1e-6,  # Genome wide significant p-value
        ...                    sign_marker_color="r",
        ...                    snp="ID",
        ...                    title="Test",
        ...                    xtick_label_set=xtick,
        ...                    xlabel="Chromosome",
        ...                    ylabel=r"$-log_{10}{(P)}$",
        ...                    sign_line_cols=["#D62728", "#2CA02C"],
        ...                    hline_kws={"linestyle": "--", "lw": 1.3},
        ...                    is_annotate_topsnp=True,
        ...                    ld_block_size=50000,  # 50000 bp
        ...                    text_kws={"fontsize": 12,  # The fontsize of text
        ...                              "arrowprops": dict(arrowstyle="-", color="k", alpha=0.6)},
        ...                    ax=ax)
    """
    if not isinstance(data, DataFrame):
        raise ValueError("[ERROR] Input data must be a pandas.DataFrame.")
    if chrom not in data:
        raise ValueError("[ERROR] Column \"%s\" not found!" % chrom)
    if pos not in data:
        raise ValueError("[ERROR] Column \"%s\" not found!" % pos)
    if pv not in data:
        raise ValueError("[ERROR] Column \"%s\" not found!" % pv)
    if is_annotate_topsnp and (snp not in data):
        raise ValueError("[ERROR] You're trying to annotate a set of SNPs but "
                         "NO SNP \"%s\" column found!" % snp)
    if CHR is not None and xtick_label_set is not None:
        raise ValueError("[ERROR] ``CHR`` and ``xtick_label_set`` can't be set simultaneously.")

    data[[chrom]] = data[[chrom]].astype(str)  # make sure all the chromosome id are character.

    # Draw the plot and return the Axes
    if ax is None:
        # ax = plt.gca()
        _, ax = subplots(figsize=(9, 3), facecolor="w", edgecolor="k")  # default

    if xticklabel_kws is None:
        xticklabel_kws = {}
    if hline_kws is None:
        hline_kws = {}
    if text_kws is None:
        text_kws = {}

    if "," in color:
        color = color.split(",")
    colors = cycle(color)

    last_xpos = 0
    xs_by_id = []  # use for collecting chromosome's position on x-axis
    x, y, c = [], [], []
    sign_snp_sites = []
    for seqid, group_data in data.groupby(by=chrom, sort=False):  # keep the raw order of chromosome

        if (CHR is not None) and (seqid != CHR):
            continue

        color = next(colors)
        for i, (site, p_value) in enumerate(zip(group_data[pos], group_data[pv])):
            y_value = -np.log10(p_value) if logp else p_value

            x.append(last_xpos + site)
            y.append(y_value)

            c.append(sign_marker_color if ((sign_marker_p is not None) and (p_value <= sign_marker_p)) else color)
            if (sign_marker_p is not None) and (p_value <= sign_marker_p):
                snp_id = group_data[snp].iloc[i]
                sign_snp_sites.append([last_xpos + site, y_value, snp_id])  # x_pos, y_value, text

        # ``xs_by_id`` is for setting up positions and ticks. Ticks should
        # be placed in the middle of a chromosome. The a new pos column is 
        # added that keeps a running sum of the positions of each successive 
        # chromsome.
        xs_by_id.append([seqid, last_xpos + (group_data[pos].iloc[0] + group_data[pos].iloc[-1]) / 2])
        last_xpos = x[-1]  # keep track so that chromosome will not overlap in the plot.

    if not x:
        raise ValueError("zero-size array to reduction operation minimum which has no "
                         "identity. This could be caused by zero-size array of ``x`` "
                         "in the ``manhattanplot(...)`` function.")

    if "marker" not in kwargs:
        kwargs["marker"] = marker

    # plot the main manhattan dot plot
    ax.scatter(x, y, c=c, alpha=alpha, edgecolors="none", **kwargs)

    if is_annotate_topsnp is not None:
        index = _find_SNPs_which_overlap_sign_neighbour_region(
            sign_snp_neighbour_region=_sign_snp_regions(sign_snp_sites, ld_block_size),
            x=x)

        # reset color for all SNPs which nearby the top SNPs.
        for i in index:
            ax.scatter(x[i], y[i], c=sign_marker_color, alpha=alpha, edgecolors="none", **kwargs)

    # Add GWAS significant lines
    if "color" in hline_kws:
        hline_kws.pop("color")

    sign_line_cols = sign_line_cols.split(",") if "," in sign_line_cols else sign_line_cols
    if suggestiveline is not None:
        ax.axhline(y=-np.log10(suggestiveline) if logp else suggestiveline, color=sign_line_cols[0], **hline_kws)
    if genomewideline is not None:
        ax.axhline(y=-np.log10(genomewideline) if logp else genomewideline, color=sign_line_cols[1], **hline_kws)

    # Plotting the Top SNP for each significant block
    if is_annotate_topsnp:
        sign_top_snp = _find_top_snp(sign_snp_sites, ld_block_size=ld_block_size, is_get_biggest=logp)
        if sign_top_snp:  # not empty
            texts = [ax.text(_x, _y, _text) for _x, _y, _text in sign_top_snp]
            adjust_text(texts, ax=ax, **text_kws)

    if CHR is None:

        if xtick_label_set is not None:
            ax.set_xticks([v for c, v in xs_by_id if c in xtick_label_set])
            ax.set_xticklabels([c for c, v in xs_by_id if c in xtick_label_set], **xticklabel_kws)
        else:
            ax.set_xticks([v for c, v in xs_by_id])
            ax.set_xticklabels([c for c, v in xs_by_id], **xticklabel_kws)

    else:
        # show the whole chromosomal position without scientific notation
        # if you are just interesting in this chromosome.
        ax.get_xaxis().get_major_formatter().set_scientific(False)

    ax.set_xlim(0, x[-1])
    ax.set_ylim(ymin=min(y), ymax=1.2 * max(y))

    if title:
        ax.set_title(title)
    if xlabel:
        ax.set_xlabel(xlabel)
    if ylabel:
        ax.set_ylabel(ylabel)

    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    return ax


def _find_top_snp(sign_snp_data, ld_block_size, is_get_biggest=True):
    """
    :param sign_snp_data:  A 2D array: [[xpos1, yvalue1, text1], [xpos2, yvalue2, text2], ...]
    """
    top_snp = []
    tmp_cube = []
    for i, (_x, _y, text) in enumerate(sign_snp_data):
        if i == 0:
            tmp_cube.append([_x, _y, text])
            continue

        if _x > tmp_cube[-1][0] + ld_block_size:
            # Sorted by y_value in increase/decrease order and only get the first value [0], which is the TopSNP.
            top_snp.append(sorted(tmp_cube, key=(lambda x: x[1]), reverse=is_get_biggest)[0])
            tmp_cube = []

        tmp_cube.append([_x, _y, text])

    if tmp_cube:  # deal the last one
        top_snp.append(sorted(tmp_cube, key=(lambda x: x[1]), reverse=True)[0])

    return top_snp


def _sign_snp_regions(sign_snp_data, ld_block_size):
    """Create region according to the coordinate of sign_snp_data."""
    regions = []
    for i, (_x, _y, _t) in enumerate(sign_snp_data):
        if i == 0:
            regions.append([_x - ld_block_size, _x])
            continue

        if _x > regions[-1][1] + ld_block_size:
            regions[-1][1] += ld_block_size
            regions.append([_x - ld_block_size, _x])
        else:
            regions[-1][1] = _x

    # The last
    if regions:
        regions[-1][1] += ld_block_size

    return regions


def _find_SNPs_which_overlap_sign_neighbour_region(sign_snp_neighbour_region, x):
    """
    """
    x_size = len(x)
    reg_size = len(sign_snp_neighbour_region)
    index = []
    tmp_index = 0
    for i in range(x_size):
        _x = x[i]

        is_overlap = False
        iter_index = range(tmp_index, reg_size)
        for j in iter_index:
            if _x > sign_snp_neighbour_region[j][1]: continue
            if _x < sign_snp_neighbour_region[j][0]: break

            tmp_index = j
            is_overlap = True
            break

        if is_overlap:
            index.append(i)

    # return the index
    return index



"""Functions to cluster data.

Author: Shujia Huang
Date: 2021-04-30 11:50:26
"""
import warnings
import numpy as np
import pandas as pd

try:
    from scipy.cluster import hierarchy

    _no_scipy = False
except ImportError:
    _no_scipy = True

from ..utils import deprecate_positional_args


class _Dendrogram(object):
    """Agglomerative hierarchical clustering by scipy.cluster.hierarchy."""

    def __init__(self, data, linkage, method, metric, axis):
        """Calculating the relationships between the columns of data.

        Parameters
        ----------
        data : pandas.DataFrame
            Rectangular data

        axis : int, optional
            Which axis to use to calculate linkage. 0 is rows, 1 is columns.
        """
        if axis == 1:
            data = data.T

        if isinstance(data, pd.DataFrame):
            array = data.values
        else:
            array = np.asarray(data)
            data = pd.DataFrame(array)

        self.array = array
        self.data = data
        self.axis = axis

        self.shape = self.data.shape
        self.metric = metric
        self.method = method

        if linkage is None:
            # call linkage function.
            self.linkage = self.calculated_linkage
        else:
            self.linkage = linkage

        self.dendrogram = self.calculate_dendrogram()
        # self.dependent_coord = self.dendrogram["dcoord"]
        # self.independent_coord = self.dendrogram["icoord"]

    def _calculate_linkage_scipy(self):
        linkage = hierarchy.linkage(self.array, method=self.method,
                                    metric=self.metric)
        return linkage

    def _calculate_linkage_fastcluster(self):
        import fastcluster
        # Fastcluster has a memory-saving vectorized version, but only
        # with certain linkage methods, and mostly with euclidean metric
        # vector_methods = ("single", "centroid", "median", "ward")
        euclidean_methods = ("centroid", "median", "ward")
        euclidean = self.metric == "euclidean" and self.method in euclidean_methods
        if euclidean or self.method == "single":
            return fastcluster.linkage_vector(self.array,
                                              method=self.method,
                                              metric=self.metric)
        else:
            linkage = fastcluster.linkage(self.array, method=self.method,
                                          metric=self.metric)
            return linkage

    @property
    def calculated_linkage(self):

        try:
            return self._calculate_linkage_fastcluster()
        except ImportError:
            if np.product(self.shape) >= 10000:
                msg = ("Clustering large matrix with scipy. Installing "
                       "`fastcluster` may give better performance.")
                warnings.warn(msg)

        return self._calculate_linkage_scipy()

    @property
    def reordered_index(self):
        """Indices of the matrix, reordered by the dendrogram"""
        return self.dendrogram["leaves"]

    def calculate_dendrogram(self):
        """Calculates a dendrogram based on the linkage matrix
        Made a separate function, not a property because don't want to
        recalculate the dendrogram every time it is accessed.
        Returns
        -------
        dendrogram : dict
            Dendrogram dictionary as returned by scipy.cluster.hierarchy
            .dendrogram. The important key-value pairing is
            "reordered_ind" which indicates the re-ordering of the matrix
        """
        return hierarchy.dendrogram(self.linkage, no_plot=True, color_threshold=-np.inf)


@deprecate_positional_args
def hierarchical_cluster(
        data=None, linkage=None, method="average", metric="euclidean", axis=1
):
    """A dendrogram cluster (or call hierarchical cluster method) to calculate
    the relationships within a matrix.

    Parameters
    ----------
    data : pandas.DaraFrame.
        Rectangular data
    linkage : numpy.array, optional
        Linkage matrix which has been calculate by other cluster method.
    method : str, optional
        Linkage method to use. Anything valid for
        scipy.cluster.hierarchy.linkage
    metric : str, optional
        Distance metric. Anything valid for scipy.spatial.distance.pdist
    axis : int, optional
        Which axis to use to calculate cluster. 0 is rows, 1 is columns.
        default: 1.

    Returns
    -------
    dendropgram : _Dendrogram
        A dendrogram object.

    """
    if _no_scipy:
        raise RuntimeError("hierarchical cluster requires scipy to be installed.")
    return _Dendrogram(data=data, linkage=linkage, method=method, metric=metric,
                       axis=axis)



"""Here is a small library for automatically adjustment of text position in matplotlib plots
to minimize overlaps. This module is based on Phlya's repository: https://github.com/Phlya/adjustText
<https://stackoverflow.com/questions/19073683/matplotlib-overlapping-annotations-text>

Modify by Shujia Huang
Date: 2021-02-05 18:39:32
"""
from itertools import product
from operator import itemgetter

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.path import get_path_collection_extents
import matplotlib


def get_bboxes_pathcollection(sc, ax):
    """Function to return a list of bounding boxes in display coordinates
    for a scatter plot
    Thank you to ImportanceOfBeingErnest:
    https://stackoverflow.com/a/55007838/1304161
    """
    #    ax.figure.canvas.draw() # need to draw before the transforms are set.
    transform = sc.get_transform()
    transOffset = sc.get_offset_transform()
    offsets = sc._offsets
    paths = sc.get_paths()
    transforms = sc.get_transforms()

    if not transform.is_affine:
        paths = [transform.transform_path_non_affine(p) for p in paths]
        transform = transform.get_affine()
    if not transOffset.is_affine:
        offsets = transOffset.transform_non_affine(offsets)
        transOffset = transOffset.get_affine()

    if isinstance(offsets, np.ma.MaskedArray):
        offsets = offsets.filled(np.nan)

    bboxes = []

    if len(paths) and len(offsets):
        if len(paths) < len(offsets):
            # for usual scatters you have one path, but several offsets
            paths = [paths[0]] * len(offsets)
        if len(transforms) < len(offsets):
            # often you may have a single scatter size, but several offsets
            transforms = [transforms[0]] * len(offsets)

        for p, o, t in zip(paths, offsets, transforms):
            result = get_path_collection_extents(
                transform.frozen(), [p], [t], [o], transOffset.frozen()
            )
            bboxes.append(result)

    return bboxes


def get_text_position(text, ax):
    x, y = text.get_position()
    x = ax.convert_xunits(x)
    y = ax.convert_yunits(y)
    t_x, t_y = text.get_transform().transform((x, y))
    return (t_x, t_y)


def set_text_position(text, t_x, t_y):
    x, y = text.get_transform().inverted().transform((t_x, t_y))
    text.set_position((x, y))
    return None


def get_orig_coords(transform, t_x, t_y):
    x, y = transform.inverted().transform((t_x, t_y))
    return (x, y)


def get_bboxes(objs, r=None, expand=(1, 1), ax=None, transform=None):
    """
    Parameters
    ----------
    objs : list, or PathCollection
        List of objects to get bboxes from. Also works with mpl PathCollection.
    r : renderer
        Renderer. The default is None, then automatically deduced from ax.
    expand : (float, float), optional
        How much to expand bboxes in (x, y), in fractions. The default is (1, 1).
    ax : Axes, optional
        The default is None, then uses current axes.
    transform : optional
        Transform to apply to the objects, if they don't return they window extent.
        The default is None, then applies the default ax transform.

    Returns
    -------
    list
        List of bboxes.

    """
    ax = ax or plt.gca()
    r = r or get_renderer(ax.get_figure())
    try:
        return [i.get_window_extent(r).expanded(*expand) for i in objs]
    except (AttributeError, TypeError):
        try:
            if all([isinstance(obj, matplotlib.transforms.BboxBase) for obj in objs]):
                return objs
            else:
                raise ValueError("Something is wrong")
        except TypeError:
            return get_bboxes_pathcollection(objs, ax)


def get_midpoint(bbox):
    cx = (bbox.x0 + bbox.x1) / 2
    cy = (bbox.y0 + bbox.y1) / 2
    return cx, cy


def get_points_inside_bbox(x, y, bbox):
    """Return the indices of points inside the given bbox."""
    x1, y1, x2, y2 = bbox.xmin, bbox.ymin, bbox.xmax, bbox.ymax
    x_in = np.logical_and(x > x1, x < x2)
    y_in = np.logical_and(y > y1, y < y2)
    return np.asarray(np.where(x_in & y_in)[0])


def get_renderer(fig):
    try:
        return fig.canvas.get_renderer()
    except AttributeError:
        return fig.canvas.renderer


def overlap_bbox_and_point(bbox, xp, yp):
    """Given a bbox that contains a given point, return the (x, y) displacement
    necessary to make the bbox not overlap the point."""
    cx, cy = get_midpoint(bbox)

    dir_x = np.sign(cx - xp)
    dir_y = np.sign(cy - yp)

    if dir_x == -1:
        dx = xp - bbox.xmax
    elif dir_x == 1:
        dx = xp - bbox.xmin
    else:
        dx = 0

    if dir_y == -1:
        dy = yp - bbox.ymax
    elif dir_y == 1:
        dy = yp - bbox.ymin
    else:
        dy = 0
    return dx, dy


def move_texts(texts, delta_x, delta_y, bboxes=None, renderer=None, ax=None):
    ax = ax or plt.gca()
    if bboxes is None:
        r = renderer or get_renderer(ax.get_figure())
        bboxes = get_bboxes(texts, r, (1, 1), ax=ax)
    ax_bbox = ax.patch.get_extents()
    xmin = ax_bbox.xmin
    xmax = ax_bbox.xmax
    ymin = ax_bbox.ymin
    ymax = ax_bbox.ymax
    for i, (text, dx, dy) in enumerate(zip(texts, delta_x, delta_y)):
        bbox = bboxes[i]
        x1, y1, x2, y2 = bbox.xmin, bbox.ymin, bbox.xmax, bbox.ymax
        if x1 + dx < xmin:
            dx = 0
        if x2 + dx > xmax:
            dx = 0
        if y1 + dy < ymin:
            dy = 0
        if y2 + dy > ymax:
            dy = 0

        x, y = get_text_position(text, ax)
        newx = x + dx
        newy = y + dy
        set_text_position(text, newx, newy)


def optimally_align_text(
        x,
        y,
        texts,
        expand=(1.0, 1.0),
        add_bboxes=None,
        renderer=None,
        ax=None,
        direction="xy",
):
    """
    For all text objects find alignment that causes the least overlap with
    points and other texts and apply it
    """
    if add_bboxes is None:
        add_bboxes = []

    ax = ax or plt.gca()
    r = renderer or get_renderer(ax.get_figure())
    ax_bbox = ax.patch.get_extents()
    xmin = ax_bbox.xmin
    xmax = ax_bbox.xmax
    ymin = ax_bbox.ymin
    ymax = ax_bbox.ymax
    bboxes = get_bboxes(texts, r, expand, ax=ax)
    if "x" not in direction:
        ha = [""]
    else:
        ha = ["left", "right", "center"]
    if "y" not in direction:
        va = [""]
    else:
        va = ["bottom", "top", "center"]
    alignment = list(product(ha, va))
    #    coords = np.array(zip(x, y))
    for i, text in enumerate(texts):
        #        tcoords = np.array(text.get_position()).T
        #        nonself_coords = coords[~np.all(coords==tcoords, axis=1)]
        #        nonself_x, nonself_y = np.split(nonself_coords, 2, axis=1)
        counts = []
        for h, v in alignment:
            if h:
                text.set_ha(h)
            if v:
                text.set_va(v)
            bbox = text.get_window_extent(r).expanded(*expand)
            c = len(get_points_inside_bbox(x, y, bbox))
            intersections = [
                bbox.intersection(bbox, bbox2) if i != j else None
                for j, bbox2 in enumerate(bboxes + add_bboxes)
            ]
            intersections = sum(
                [abs(b.width * b.height) if b is not None else 0 for b in intersections]
            )
            # Check for out-of-axes position
            bbox = text.get_window_extent(r)
            x1, y1, x2, y2 = bbox.xmin, bbox.ymin, bbox.xmax, bbox.ymax
            if x1 < xmin or x2 > xmax or y1 < ymin or y2 > ymax:
                axout = 1
            else:
                axout = 0
            counts.append((axout, c, intersections))
        # Most important: prefer alignments that keep the text inside the axes.
        # If tied, take the alignments that minimize the number of x, y points
        # contained inside the text.
        # Break any remaining ties by minimizing the total area of intersections
        # with all text bboxes and other objects to avoid.
        a, value = min(enumerate(counts), key=itemgetter(1))
        if "x" in direction:
            text.set_ha(alignment[a][0])
        if "y" in direction:
            text.set_va(alignment[a][1])
        bboxes[i] = text.get_window_extent(r).expanded(*expand)
    return texts


def repel_text(
        texts, renderer=None, ax=None, expand=(1.2, 1.2), only_use_max_min=False, move=False
):
    """
    Repel texts from each other while expanding their bounding boxes by expand
    (x, y), e.g. (1.2, 1.2) would multiply width and height by 1.2.
    Requires a renderer to get the actual sizes of the text, and to that end
    either one needs to be directly provided, or the axes have to be specified,
    and the renderer is then got from the axes object.
    """
    ax = ax or plt.gca()
    r = renderer or get_renderer(ax.get_figure())
    bboxes = get_bboxes(texts, r, expand, ax=ax)
    xmins = [bbox.xmin for bbox in bboxes]
    xmaxs = [bbox.xmax for bbox in bboxes]
    ymaxs = [bbox.ymax for bbox in bboxes]
    ymins = [bbox.ymin for bbox in bboxes]

    overlaps_x = np.zeros((len(bboxes), len(bboxes)))
    overlaps_y = np.zeros_like(overlaps_x)
    overlap_directions_x = np.zeros_like(overlaps_x)
    overlap_directions_y = np.zeros_like(overlaps_y)
    for i, bbox1 in enumerate(bboxes):
        overlaps = get_points_inside_bbox(
            xmins * 2 + xmaxs * 2, (ymins + ymaxs) * 2, bbox1
        ) % len(bboxes)
        overlaps = np.unique(overlaps)
        for j in overlaps:
            bbox2 = bboxes[j]
            x, y = bbox1.intersection(bbox1, bbox2).size
            overlaps_x[i, j] = x
            overlaps_y[i, j] = y
            direction = np.sign(bbox1.extents - bbox2.extents)[:2]
            overlap_directions_x[i, j] = direction[0]
            overlap_directions_y[i, j] = direction[1]

    move_x = overlaps_x * overlap_directions_x
    move_y = overlaps_y * overlap_directions_y

    delta_x = move_x.sum(axis=1)
    delta_y = move_y.sum(axis=1)

    q = np.sum(overlaps_x), np.sum(overlaps_y)
    if move:
        move_texts(texts, delta_x, delta_y, bboxes, ax=ax)
    return delta_x, delta_y, q


def repel_text_from_bboxes(
        add_bboxes,
        texts,
        renderer=None,
        ax=None,
        expand=(1.2, 1.2),
        only_use_max_min=False,
        move=False,
):
    """
    Repel texts from other objects' bboxes while expanding their (texts')
    bounding boxes by expand (x, y), e.g. (1.2, 1.2) would multiply width and
    height by 1.2.
    Requires a renderer to get the actual sizes of the text, and to that end
    either one needs to be directly provided, or the axes have to be specified,
    and the renderer is then got from the axes object.
    """
    ax = ax or plt.gca()
    r = renderer or get_renderer(ax.get_figure())

    bboxes = get_bboxes(texts, r, expand, ax=ax)

    overlaps_x = np.zeros((len(bboxes), len(add_bboxes)))
    overlaps_y = np.zeros_like(overlaps_x)
    overlap_directions_x = np.zeros_like(overlaps_x)
    overlap_directions_y = np.zeros_like(overlaps_y)

    for i, bbox1 in enumerate(bboxes):
        for j, bbox2 in enumerate(add_bboxes):
            try:
                x, y = bbox1.intersection(bbox1, bbox2).size
                direction = np.sign(bbox1.extents - bbox2.extents)[:2]
                overlaps_x[i, j] = x
                overlaps_y[i, j] = y
                overlap_directions_x[i, j] = direction[0]
                overlap_directions_y[i, j] = direction[1]
            except AttributeError:
                pass

    move_x = overlaps_x * overlap_directions_x
    move_y = overlaps_y * overlap_directions_y

    delta_x = move_x.sum(axis=1)
    delta_y = move_y.sum(axis=1)

    q = np.sum(overlaps_x), np.sum(overlaps_y)
    if move:
        move_texts(texts, delta_x, delta_y, bboxes, ax=ax)
    return delta_x, delta_y, q


def repel_text_from_points(x, y, texts, renderer=None, ax=None, expand=(1.2, 1.2), move=False):
    """
    Repel texts from all points specified by x and y while expanding their
    (texts'!) bounding boxes by expandby  (x, y), e.g. (1.2, 1.2)
    would multiply both width and height by 1.2.
    Requires a renderer to get the actual sizes of the text, and to that end
    either one needs to be directly provided, or the axes have to be specified,
    and the renderer is then got from the axes object.
    """
    assert len(x) == len(y)
    ax = ax or plt.gca()
    r = renderer or get_renderer(ax.get_figure())
    bboxes = get_bboxes(texts, r, expand, ax=ax)

    # move_x[i,j] is the x displacement of the i'th text caused by the j'th point
    move_x = np.zeros((len(bboxes), len(x)))
    move_y = np.zeros((len(bboxes), len(x)))
    for i, bbox in enumerate(bboxes):
        xy_in = get_points_inside_bbox(x, y, bbox)
        for j in xy_in:
            xp, yp = x[j], y[j]
            dx, dy = overlap_bbox_and_point(bbox, xp, yp)

            move_x[i, j] = dx
            move_y[i, j] = dy

    delta_x = move_x.sum(axis=1)
    delta_y = move_y.sum(axis=1)
    q = np.sum(np.abs(move_x)), np.sum(np.abs(move_y))
    if move:
        move_texts(texts, delta_x, delta_y, bboxes, ax=ax)
    return delta_x, delta_y, q


def repel_text_from_axes(texts, ax=None, bboxes=None, renderer=None, expand=None):
    ax = ax or plt.gca()
    r = renderer or get_renderer(ax.get_figure())
    if expand is None:
        expand = (1, 1)
    if bboxes is None:
        bboxes = get_bboxes(texts, r, expand=expand, ax=ax)
    ax_bbox = ax.patch.get_extents()
    xmin = ax_bbox.xmin
    xmax = ax_bbox.xmax
    ymin = ax_bbox.ymin
    ymax = ax_bbox.ymax
    for i, bbox in enumerate(bboxes):
        x1, y1, x2, y2 = bbox.xmin, bbox.ymin, bbox.xmax, bbox.ymax
        dx, dy = 0, 0
        if x1 < xmin:
            dx = xmin - x1
        if x2 > xmax:
            dx = xmax - x2
        if y1 < ymin:
            dy = ymin - y1
        if y2 > ymax:
            dy = ymax - y2
        if dx or dy:
            x, y = get_text_position(texts[i], ax)
            newx, newy = x + dx, y + dy
            set_text_position(texts[i], newx, newy)
    return texts


def float_to_tuple(a):
    try:
        a = float(a)
        return (a, a)
    except TypeError:
        assert len(a) == 2
        try:
            b = float(a[0]), float(a[1])
        except TypeError:
            raise TypeError("Force values must be castable to floats")
        return b


def adjust_text(
        texts,
        x=None,
        y=None,
        add_objects=None,
        ax=None,
        expand_text=(1.05, 1.2),
        expand_points=(1.05, 1.2),
        expand_objects=(1.05, 1.2),
        expand_align=(1.05, 1.2),
        autoalign="xy",
        va="center",
        ha="center",
        force_text=(0.1, 0.25),
        force_points=(0.2, 0.5),
        force_objects=(0.1, 0.25),
        lim=500,
        precision=0.01,
        only_move=None,
        avoid_text=True,
        avoid_points=True,
        avoid_self=True,
        save_steps=False,
        save_prefix="",
        save_format="png",
        add_step_numbers=True,
        *args,
        **kwargs):
    """Iteratively adjusts the locations of texts.

    Call adjust_text the very last, after all plotting (especially
    anything that can change the axes limits) has been done. This is
    because to move texts the function needs to use the dimensions of
    the axes, and without knowing the final size of the plots the
    results will be completely nonsensical, or suboptimal.

    First moves all texts that are outside the axes limits
    inside. Then in each iteration moves all texts away from each
    other and from points. In the end hides texts and substitutes them
    with annotations to link them to the respective points.

    Parameters
    ----------
    texts : list
        A list of :obj:`matplotlib.text.Text` objects to adjust.

    Other Parameters
    ----------------
    x : array_like
        x-coordinates of points to repel from; if not provided only uses text
        coordinates.
    y : array_like
        y-coordinates of points to repel from; if not provided only uses text
        coordinates
    add_objects : list or PathCollection
        a list of additional matplotlib objects to avoid; they must have a
        `.get_window_extent()` method; alternatively, a PathCollection or a
        list of Bbox objects.
    ax : matplotlib axe, default is current axe (plt.gca())
        axe object with the plot
    expand_text : array_like, default (1.05, 1.2)
        a tuple/list/... with 2 multipliers (x, y) by which to expand the
        bounding box of texts when repelling them from each other.
    expand_points : array_like, default (1.05, 1.2)
        a tuple/list/... with 2 multipliers (x, y) by which to expand the
        bounding box of texts when repelling them from points.
    expand_objects : array_like, default (1.05, 1.2)
        a tuple/list/... with 2 multipliers (x, y) by which to expand the
        bounding box of texts when repelling them from other objects.
    expand_align : array_like, default (1.05, 1.2)
        a tuple/list/... with 2 multipliers (x, y) by which to expand the
        bounding box of texts when autoaligning texts.
    autoalign: str or boolean {'xy', 'x', 'y', True, False}, default 'xy'
        Direction in wich the best alignement will be determined

        - 'xy' or True, best alignment of all texts determined in all
          directions automatically before running the iterative adjustment
          (overriding va and ha),
        - 'x', will only align horizontally,
        - 'y', will only align vertically,
        - False, do nothing (i.e. preserve va and ha)

    va : str, default 'center'
        vertical alignment of texts
    ha : str, default 'center'
        horizontal alignment of texts,
    force_text : tuple, default (0.1, 0.25)
        the repel force from texts is multiplied by this value
    force_points : tuple, default (0.2, 0.5)
        the repel force from points is multiplied by this value
    force_objects : float, default (0.1, 0.25)
        same as other forces, but for repelling additional objects
    lim : int, default 500
        limit of number of iterations
    precision : float, default 0.01
        iterate until the sum of all overlaps along both x and y are less than
        this amount, as a fraction of the total widths and heights,
        respectively. May need to increase for complicated situations.
    only_move : dict, default {'points':'xy', 'text':'xy', 'objects':'xy'}
        a dict to restrict movement of texts to only certain axes for certain
        types of overlaps.
        Valid keys are 'points', 'text', and 'objects'.
        Valid values are '', 'x', 'y', and 'xy'.
        For example, only_move={'points':'y', 'text':'xy', 'objects':'xy'}
        forbids moving texts along the x axis due to overlaps with points.
    avoid_text : bool, default True
        whether to repel texts from each other.
    avoid_points : bool, default True
        whether to repel texts from points. Can be helpful to switch off in
        extremely crowded plots.
    avoid_self : bool, default True
        whether to repel texts from its original positions.
    save_steps : bool, default False
        whether to save intermediate steps as images.
    save_prefix : str, default ''
        if `save_steps` is True, a path and/or prefix to the saved steps.
    save_format : str, default 'png'
        if `save_steps` is True, a format to save the steps into.
    add_step_numbers : bool, default True
        if `save_steps` is True, whether to add step numbers as titles to the
        images of saving steps.
    args and kwargs :
        any arguments will be fed into obj:`ax.annotate` after all the
        optimization is done just for plotting the connecting arrows if
        required.

    Return
    ------
    int
        Number of iteration
    """
    if only_move is None:
        only_move = {"points": "xy", "text": "xy", "objects": "xy"}

    plt.draw()
    ax = ax or plt.gca()
    r = get_renderer(ax.get_figure())
    transform = texts[0].get_transform()
    if (x is not None) & (y is not None):
        for ix, tupxy in enumerate(zip(x, y)):
            t_x, t_y = transform.transform(tupxy)
            x[ix] = t_x
            y[ix] = t_y

    orig_xy = [get_text_position(text, ax) for text in texts]
    orig_x = [xy[0] for xy in orig_xy]
    orig_y = [xy[1] for xy in orig_xy]
    force_objects = float_to_tuple(force_objects)
    force_text = float_to_tuple(force_text)
    force_points = float_to_tuple(force_points)

    #    xdiff = np.diff(ax.get_xlim())[0]
    #    ydiff = np.diff(ax.get_ylim())[0]

    bboxes = get_bboxes(texts, r, (1.0, 1.0), ax)
    sum_width = np.sum(list(map(lambda bbox: bbox.width, bboxes)))
    sum_height = np.sum(list(map(lambda bbox: bbox.height, bboxes)))
    if not any(list(map(lambda val: "x" in val, only_move.values()))):
        precision_x = np.inf
    else:
        precision_x = precision * sum_width

    if not any(list(map(lambda val: "y" in val, only_move.values()))):
        precision_y = np.inf
    else:
        precision_y = precision * sum_height

    if x is None:
        if y is None:
            if avoid_self:
                x, y = orig_x, orig_y
            else:
                x, y = [], []
        else:
            raise ValueError("Please specify both x and y, or neither")
    if y is None:
        raise ValueError("Please specify both x and y, or neither")
    if add_objects is None:
        text_from_objects = False
        add_bboxes = []
    else:
        try:
            add_bboxes = get_bboxes(add_objects, r, (1, 1), ax)
        except:
            raise ValueError("Can't get bounding boxes from add_objects - is it a "
                             "flat list of matplotlib objects?")

        text_from_objects = True

    for text in texts:
        text.set_va(va)
        text.set_ha(ha)

    if save_steps:
        if add_step_numbers:
            plt.title("Before")
        plt.savefig("%s%s.%s" % (save_prefix, "000a", save_format), format=save_format, dpi=150)

    if autoalign:
        if autoalign is True:
            autoalign = "xy"
        for i in range(2):
            texts = optimally_align_text(
                x,
                y,
                texts,
                expand=expand_align,
                add_bboxes=add_bboxes,
                direction=autoalign,
                renderer=r,
                ax=ax,
            )

    if save_steps:
        if add_step_numbers:
            plt.title("Autoaligned")
        plt.savefig(
            "%s%s.%s" % (save_prefix, "000b", save_format), format=save_format, dpi=150
        )

    texts = repel_text_from_axes(texts, ax, renderer=r, expand=expand_points)
    history = [(np.inf, np.inf)] * 10
    for i in range(lim):
        #        q1, q2 = [np.inf, np.inf], [np.inf, np.inf]

        if avoid_text:
            d_x_text, d_y_text, q1 = repel_text(
                texts, renderer=r, ax=ax, expand=expand_text
            )
        else:
            d_x_text, d_y_text, q1 = [0] * len(texts), [0] * len(texts), (0, 0)

        if avoid_points:
            d_x_points, d_y_points, q2 = repel_text_from_points(
                x, y, texts, ax=ax, renderer=r, expand=expand_points
            )
        else:
            d_x_points, d_y_points, q2 = [0] * len(texts), [0] * len(texts), (0, 0)

        if text_from_objects:
            d_x_objects, d_y_objects, q3 = repel_text_from_bboxes(
                add_bboxes, texts, ax=ax, renderer=r, expand=expand_objects
            )
        else:
            d_x_objects, d_y_objects, q3 = [0] * len(texts), [0] * len(texts), (0, 0)

        if only_move:
            if "text" in only_move:
                if "x" not in only_move["text"]:
                    d_x_text = np.zeros_like(d_x_text)
                if "y" not in only_move["text"]:
                    d_y_text = np.zeros_like(d_y_text)
            if "points" in only_move:
                if "x" not in only_move["points"]:
                    d_x_points = np.zeros_like(d_x_points)
                if "y" not in only_move["points"]:
                    d_y_points = np.zeros_like(d_y_points)
            if "objects" in only_move:
                if "x" not in only_move["objects"]:
                    d_x_objects = np.zeros_like(d_x_objects)
                if "y" not in only_move["objects"]:
                    d_y_objects = np.zeros_like(d_y_objects)

        dx = (
                np.array(d_x_text) * force_text[0]
                + np.array(d_x_points) * force_points[0]
                + np.array(d_x_objects) * force_objects[0]
        )
        dy = (
                np.array(d_y_text) * force_text[1]
                + np.array(d_y_points) * force_points[1]
                + np.array(d_y_objects) * force_objects[1]
        )
        qx = np.sum([q[0] for q in [q1, q2, q3]])
        qy = np.sum([q[1] for q in [q1, q2, q3]])
        histm = np.max(np.array(history), axis=0)
        history.pop(0)
        history.append((qx, qy))
        move_texts(texts, dx, dy, bboxes=get_bboxes(texts, r, (1, 1), ax), ax=ax)
        if save_steps:
            if add_step_numbers:
                plt.title(i + 1)
            plt.savefig(
                "%s%s.%s" % (save_prefix, "{0:03}".format(i + 1), save_format),
                format=save_format,
                dpi=150,
            )
        # Stop if we've reached the precision threshold, or if the x and y displacement
        # are both greater than the max over the last 10 iterations (suggesting a
        # failure to converge)
        if (qx < precision_x and qy < precision_y) or np.all([qx, qy] >= histm):
            break

    # Now adding arrows from texts to their original locations if required
    if "arrowprops" in kwargs:
        bboxes = get_bboxes(texts, r, (1, 1), ax)
        kwap = kwargs.pop("arrowprops")
        for j, (bbox, text) in enumerate(zip(bboxes, texts)):
            ap = {"patchA": text}  # Ensure arrow is clipped by the text
            ap.update(kwap)  # Add arrowprops from kwargs
            ax.annotate(
                "",  # Add an arrow from the text to the point
                xy=get_orig_coords(transform, orig_x[j], orig_y[j]),
                xytext=transform.inverted().transform(get_midpoint(bbox)),
                arrowprops=ap,
                xycoords=transform,
                textcoords=transform,
                *args,
                **kwargs
            )

    if save_steps:
        if add_step_numbers:
            plt.title(i + 1)
            plt.savefig(
                "%s%s.%s" % (save_prefix, "{0:03}".format(i + 1), save_format),
                format=save_format,
                dpi=150)

    return i+1



""" plot 2-6 sets venn diagram

Author: Shujia Huang
Date: 2021-04-26 12:03:00

Thanks to the code from tctianchi and LankyCyril: https://github.com/tctianchi/pyvenn
"""
from warnings import warn

from matplotlib.pyplot import subplots
from matplotlib.colors import to_rgba
from matplotlib.patches import Ellipse, Polygon

# from matplotlib.cm import ScalarMappable
from ..palette import generate_colors_palette

__all__ = ["venn", "vennx", "generate_petal_labels"]

SHAPE_COORDS = {
    2: [(.375, .500), (.625, .500)],
    3: [(.333, .633), (.666, .633), (.500, .310)],
    4: [(.350, .400), (.450, .500), (.544, .500), (.644, .400)],
    5: [(.428, .449), (.469, .543), (.558, .523), (.578, .432), (.489, .383)],
    6: [(.637, .921, .649, .274, .188, .667),
        (.981, .769, .335, .191, .393, .671),
        (.941, .397, .292, .475, .456, .747),
        (.662, .119, .316, .548, .662, .700),
        (.309, .081, .374, .718, .681, .488),
        (.016, .626, .726, .687, .522, .327)]
}

SHAPE_DIMS = {
    2: [(.50, .50), (.50, .50)],
    3: [(.50, .50), (.50, .50), (.50, .50)],
    4: [(.72, .45), (.72, .45), (.72, .45), (.72, .45)],
    5: [(.87, .50), (.87, .50), (.87, .50), (.87, .50), (.87, .50)],
    6: [(None,)] * 6
}

SHAPE_ANGLES = {
    2: [0, 0],
    3: [0, 0, 0],
    4: [140, 140, 40, 40],
    5: [155, 82, 10, 118, 46],
    6: [None] * 6
}

PETAL_LABEL_COORDS = {
    2: {"01": (.74, .50), "10": (.26, .50), "11": (.50, .50)},
    3: {"001": (.500, .270), "010": (.730, .650), "011": (.610, .460),
        "100": (.270, .650), "101": (.390, .460), "110": (.500, .650),
        "111": (.500, .508)},
    4: {"0001": (.85, .42), "0010": (.68, .72), "0011": (.77, .59),
        "0100": (.32, .72), "0101": (.71, .30), "0110": (.50, .66),
        "0111": (.65, .50), "1000": (.14, .42), "1001": (.50, .17),
        "1010": (.29, .30), "1011": (.39, .24), "1100": (.23, .59),
        "1101": (.61, .24), "1110": (.35, .50), "1111": (.50, .38)},
    5: {"00001": (.27, .11), "00010": (.72, .11), "00011": (.55, .13),
        "00100": (.91, .58), "00101": (.78, .64), "00110": (.84, .41),
        "00111": (.76, .55), "01000": (.51, .90), "01001": (.39, .15),
        "01010": (.42, .78), "01011": (.50, .15), "01100": (.67, .76),
        "01101": (.70, .71), "01110": (.51, .74), "01111": (.64, .67),
        "10000": (.10, .61), "10001": (.20, .31), "10010": (.76, .25),
        "10011": (.65, .23), "10100": (.18, .50), "10101": (.21, .37),
        "10110": (.81, .37), "10111": (.74, .40), "11000": (.27, .70),
        "11001": (.34, .25), "11010": (.33, .72), "11011": (.51, .22),
        "11100": (.25, .58), "11101": (.28, .39), "11110": (.36, .66),
        "11111": (.51, .47)},
    6: {"000001": (.212, .562), "000010": (.430, .249), "000011": (.356, .444),
        "000100": (.609, .255), "000101": (.323, .546), "000110": (.513, .316),
        "000111": (.523, .348), "001000": (.747, .458), "001001": (.325, .492),
        "001010": (.670, .481), "001011": (.359, .478), "001100": (.653, .444),
        "001101": (.344, .526), "001110": (.653, .466), "001111": (.363, .503),
        "010000": (.750, .616), "010001": (.682, .654), "010010": (.402, .310),
        "010011": (.392, .421), "010100": (.653, .691), "010101": (.651, .644),
        "010110": (.490, .340), "010111": (.468, .399), "011000": (.692, .545),
        "011001": (.666, .592), "011010": (.665, .496), "011011": (.374, .470),
        "011100": (.653, .537), "011101": (.652, .579), "011110": (.653, .488),
        "011111": (.389, .486), "100000": (.553, .806), "100001": (.313, .604),
        "100010": (.388, .694), "100011": (.375, .633), "100100": (.605, .359),
        "100101": (.334, .555), "100110": (.582, .397), "100111": (.542, .372),
        "101000": (.468, .708), "101001": (.355, .572), "101010": (.420, .679),
        "101011": (.375, .597), "101100": (.641, .436), "101101": (.348, .538),
        "101110": (.635, .453), "101111": (.370, .548), "110000": (.594, .689),
        "110001": (.579, .670), "110010": (.398, .670), "110011": (.395, .653),
        "110100": (.633, .682), "110101": (.616, .656), "110110": (.587, .427),
        "110111": (.526, .415), "111000": (.495, .677), "111001": (.505, .648),
        "111010": (.428, .663), "111011": (.430, .631), "111100": (.639, .524),
        "111101": (.591, .604), "111110": (.622, .477), "111111": (.501, .523)}
}

DATASET_LEGEND_COORDS = {
    # n_set => [x, y, horizontalalignment, verticalalignment]
    2: {
        0: (0.20, 0.76, "right", "bottom"),
        1: (0.80, 0.76, "left", "bottom"),
    },
    3: {
        0: (0.15, 0.87, "right", "bottom"),
        1: (0.85, 0.87, "left", "bottom"),
        2: (0.50, 0.02, "center", "top"),
    },
    4: {
        0: (0.13, 0.18, "right", "center"),
        1: (0.18, 0.83, "right", "bottom"),
        2: (0.82, 0.83, "left", "bottom"),
        3: (0.87, 0.18, "left", "top"),
    },
    5: {
        0: (0.02, 0.72, "right", "center"),
        1: (0.72, 0.94, "center", "bottom"),
        2: (0.97, 0.74, "left", "center"),
        3: (0.88, 0.05, "left", "center"),
        4: (0.12, 0.05, "right", "center"),
    },
    6: {
        0: (0.674, 0.824, "center", "center"),
        1: (0.747, 0.751, "center", "center"),
        2: (0.739, 0.396, "center", "center"),
        3: (0.700, 0.247, "center", "center"),
        4: (0.291, 0.255, "center", "center"),
        5: (0.203, 0.484, "center", "center"),
    }
}


def generate_colors(cmap="viridis", n_colors=6, alpha=.4):
    """Generate colors from matplotlib colormap; pass list to use exact colors"""
    if not isinstance(n_colors, int) or (n_colors < 2) or (n_colors > 6):
        raise ValueError("n_colors must be an integer between 2 and 6")

    colors = generate_colors_palette(cmap=cmap, n_colors=n_colors, alpha=alpha)
    return colors[:n_colors]


def less_transparent_color(color, alpha_factor=2):
    """Bump up color's alpha"""
    new_alpha = (1 + to_rgba(color)[3]) / alpha_factor
    return to_rgba(color, alpha=new_alpha)


def draw_ellipse(ax, x, y, w, h, a, color):
    """Wrapper for drawing ellipse; called like `draw_ellipse(ax, *coords, *dims, angle, color)`"""
    ax.add_patch(
        Ellipse(
            xy=(x, y),
            width=w,
            height=h,
            angle=a,
            facecolor=color,
            edgecolor=less_transparent_color(color),
            lw=1,
        )
    )


def draw_triangle(ax, x1, y1, x2, y2, x3, y3, _dim, _angle, color):
    """Wrapper for drawing triangle; called like `draw_triangle(ax, *coords, None, None, color)`"""
    ax.add_patch(
        Polygon(
            xy=[(x1, y1), (x2, y2), (x3, y3)],
            closed=True,
            facecolor=color,
            edgecolor=less_transparent_color(color),
            lw=1,
        )
    )
    return


def draw_text(ax, x, y, text, color="black", fontsize=14, ha="center", va="center"):
    """Wrapper for drawing text"""
    ax.text(
        x, y, text,
        fontsize=fontsize,
        color=color,
        horizontalalignment=ha,
        verticalalignment=va,
    )
    return


def _generate_logics(n_sets):
    """Generate intersection identifiers in binary (0010 etc)"""
    for i in range(1, 2 ** n_sets):
        yield bin(i)[2:].zfill(n_sets)


def generate_petal_labels(datasets, fmt="{size}"):
    """Generate petal descriptions for venn diagram based on set sizes."""
    datasets = list(datasets)
    n_sets = len(datasets)
    if n_sets < 2 or n_sets > 6:
        raise ValueError("Number of sets must be between 2 and 6.")

    dataset_union = set.union(*datasets)
    universe_size = len(dataset_union)
    petal_labels = {}
    for logic in _generate_logics(n_sets):
        included_sets = [datasets[i] for i in range(n_sets) if logic[i] == "1"]
        excluded_sets = [datasets[i] for i in range(n_sets) if logic[i] == "0"]
        petal_set = (
                (dataset_union & set.intersection(*included_sets)) -
                set.union(set(), *excluded_sets)
        )
        petal_labels[logic] = fmt.format(
            logic=logic,
            size=len(petal_set),
            percentage=(100 * len(petal_set) / max(universe_size, 1))
        )
    return petal_labels


def get_labels(data, fill=None):
    """
    get a dict of labels for groups in data

    @type data: list[Iterable]
    @rtype: dict[str, str]

    input
      data: data to get label for
      fill: ["number"|"logic"|"percent"]

    return
      labels: a dict of labels for different sets

    example:
    In [12]: get_labels([range(10), range(5,15), range(3,8)], fill=["number"])
    Out[12]:
    {'001': '0',
     '010': '5',
     '011': '0',
     '100': '3',
     '101': '2',
     '110': '2',
     '111': '3'}
    """
    warn((
            "`get_labels()` is retained for backwards compatibility; " +
            "use `generate_petal_labels()` or the higher level `venn()` instead"
    ))
    if fill is None:
        fill = ["number"]

    fmt = ""
    if "logic" in fill:
        fmt += "{logic}: "
    if "number" in fill:
        fmt += "{size} "
    if "percent" in fill:
        fmt += "({percentage:.1f}%)"
    return generate_petal_labels(data, fmt)


def _init_axes(ax):
    """Create axes if do not exist, set axes parameters"""
    if ax is None:
        _, ax = subplots(nrows=1, ncols=1, figsize=(7, 7))

    ax.set_axis_off()
    return ax


def _get_n_sets(petal_labels, dataset_labels):
    """Infer number of sets, check consistency"""
    n_sets = len(dataset_labels)
    petal_labels_set = set(_generate_logics(n_sets))
    for logic in petal_labels.keys():
        if len(logic) != n_sets:
            raise ValueError("Inconsistent petal and dataset labels: %d, %d" % (len(logic), n_sets))
        if not (set(logic) <= {"0", "1"}):
            raise KeyError("Key not understood: " + logic)
        if logic not in petal_labels_set:
            raise KeyError("'%s' is not a legal key." % logic)
    return n_sets


def _draw_venn(data, names=None, palette=None, alpha=0.4, fontsize=14,
               legend_use_petal_color=False, legend_loc=None, ax=None):
    """Draw Venn diagram, annotate petals and dataset labels.
    """
    DEFAULT_COLORS = [
        # r, g, b, a
        [0.361, 0.753, 0.384, 0.5],
        [0.353, 0.608, 0.831, 0.5],
        [0.965, 0.925, 0.337, 0.6],
        [0.945, 0.353, 0.376, 0.4],
        [1.000, 0.459, 0.000, 0.3],
        [0.322, 0.322, 0.745, 0.2]
    ]

    if not isinstance(names, list):
        raise ValueError("Names of sets should be a list and must not be empty.")

    n_sets = _get_n_sets(data, names)
    if 2 <= n_sets < 6:
        draw_shape = draw_ellipse
    elif n_sets == 6:
        draw_shape = draw_triangle
    else:
        raise ValueError("Number of sets must be between 2 and 6")

    ax = _init_axes(ax)
    if palette is None:
        palette = [DEFAULT_COLORS[i] for i in range(n_sets)]
    else:
        palette = generate_colors(n_colors=n_sets, cmap=palette, alpha=alpha)

    shape_params = zip(
        SHAPE_COORDS[n_sets],
        SHAPE_DIMS[n_sets],
        SHAPE_ANGLES[n_sets],
        palette
    )

    # Draw the shape for venn diagram
    for coords, dims, angle, color in shape_params:
        draw_shape(ax, *coords, *dims, angle, color)

    # annotate the value for each petal of venn plot
    for k, value in data.items():
        if k in PETAL_LABEL_COORDS[n_sets]:
            x, y = PETAL_LABEL_COORDS[n_sets][k]
            draw_text(ax, x, y, value, fontsize=fontsize)

    if legend_loc is not None:
        ax.legend(names, loc=legend_loc, prop={"size": fontsize})
    else:
        # plot the legend name for each dataset
        for i in range(n_sets):
            x, y, ha, va = DATASET_LEGEND_COORDS[n_sets][i]
            if legend_use_petal_color:
                c = palette[i]
                c[-1] = 1.0  # Set alpha blending value to opaque for legend text.
            else:
                c = "k"
            draw_text(ax, x, y, names[i], fontsize=fontsize + 2, color=c, ha=ha, va=va)

    return ax


def is_valid_dataset_dict(data):
    """Validate passed data (must be dictionary of sets)"""
    if not (hasattr(data, "keys") and hasattr(data, "values")):
        return False
    for dataset in data.values():
        if not isinstance(dataset, set):
            return False
    else:
        return True


def is_already_venn_dataset(petal_labels, dataset_labels):
    if not isinstance(dataset_labels, list):
        return False

    if not (hasattr(petal_labels, "keys") and hasattr(petal_labels, "values")):
        return False

    n_sets = len(dataset_labels)
    petal_labels_set = set(_generate_logics(n_sets))
    valid = True
    for logic in petal_labels.keys():
        if not isinstance(petal_labels[logic], str):
            valid = False
        if len(logic) != n_sets:
            valid = False
        if not (set(logic) <= {"0", "1"}):
            valid = False
        if logic not in petal_labels_set:
            valid = False
        if not valid:
            break

    return valid


def vennx(data, names=None, palette=None, alpha=0.4, fontsize=14,
          legend_use_petal_color=False, legend_loc=None, ax=None):
    """Generate venn diagram by input petal labels data.

    Parameters
    ----------
    data : dict, require
        A dictionaries of labels for dataset that will produce the venn diagram.
        The ``data`` variable should looks like:
        {'001': '0',
         '010': '5',
         '011': '0',
         '100': '3',
         '101': '2',
         '110': '2',
         '111': '3'}

    names : list, optional, default ``list(data.keys())``
        The label names for each petal in venn plot.

    palette : string, list, or :class:`matplotlib.colors.Colormap`, optional, default: DEFAULT_COLOR.
        String values are passed to :func:`generate_colors`. List values imply categorical
        mapping, while a colormap object implies numeric mapping.

    alpha : float scalar, default is 0.4, optional
        The alpha blending value, between 0(transparent) and 1(opaque)

    fontsize : integer, optional, default: 14
        Set the fontsize for plot.

    legend_use_petal_color : bool, optional, default: False
        Determine to use the same color as petals for legend or not.

    legend_loc : string. optional
        Place a legend on the axes. String value is passed to matplotlib :rc:`legend.loc`.

    ax : matplotlib axis, optional
        Axis to plot on, otherwise create a default axis by plt.subplots() with figsize=(7, 7)
        in ``_draw_venn()``.

    Returns
    -------
    ax : matplotlib Axes
        Axes object with the venn plot.

    Examples
    --------

    Plot a three sets venn diagram.

    .. plot::
        :context: close-figs

        >>> from numpy.random import choice
        >>> from geneview.baseplot._venn import vennx, generate_petal_labels
        >>> dataset_dict = {name:set(choice(1000, 250, replace=False)) for name in list("ABCD")}
        >>> petal_labels = generate_petal_labels(dataset_dict.values(), fmt="{size}\\n({percentage:.1f}%)")
        >>> ax = vennx(data=petal_labels, names=list(dataset_dict.keys()))

    Set the labels for each petal by the keys of dataset.

    .. plot::
        :context: close-figs
        >>> ax = vennx(data=petal_labels, names=list(dataset_dict.keys()))

    Or set the labels by customer.

    .. plot::
        :context: close-figs
        >>> ax = vennx(data=petal_labels, names=["set 1", "set 2", "set 3", "set 4"],
        ...            legend_use_petal_color=True)
    """
    if not isinstance(names, list):
        raise ValueError("Names of sets should be a list and must not be empty.")

    if not is_already_venn_dataset(data, names):
        raise TypeError("``data`` is not a dict or the value is not a string. ")

    return _draw_venn(data=data,
                      names=names,
                      palette=palette,
                      alpha=alpha,
                      fontsize=fontsize,
                      legend_use_petal_color=legend_use_petal_color,
                      legend_loc=legend_loc,
                      ax=ax)


def venn(data, names=None, fmt="{size}", palette="viridis", alpha=0.4, fontsize=14,
         legend_use_petal_color=False, legend_loc=None, ax=None):
    """Check input, generate petal labels, draw venn diagram.

    Parameters
    ----------
    data : dict, require
        A dictionaries of dataset that will produce the venn diagram.

    names : list, optional, default ``list(data.keys())``
        The label names for each petal in venn plot.

    fmt : string, optional, default: "{size}"
        A Python 3 style format string that understands {size}, {percentage}, and {logic}.
        Here set the data format for petal datasets' lebals.

    palette : string, list, or :class:`matplotlib.colors.Colormap`, optional, default: "viridis".
        String values are passed to :func:`generate_colors`. List values imply categorical
        mapping, while a colormap object implies numeric mapping.

    alpha : float scalar, default is 0.4, optional
        The alpha blending value, between 0(transparent) and 1(opaque)

    fontsize : integer, optional, default: 14
        Set the fontsize for plot.

    legend_use_petal_color : bool, optional, default: False
        Determine to use the same color as petals for legend or not.

    legend_loc : string. optional
        Place a legend on the axes. String value is passed to matplotlib :rc:`legend.loc`.

    ax : matplotlib axis, optional
        Axis to plot on, otherwise create a default axis by plt.subplots() with figsize=(7, 7)
        in ``_draw_venn()``.

    Returns
    -------
    ax : matplotlib Axes
        Axes object with the venn plot.

    Examples
    --------

    Plot a minimal venn plot example.

    .. plot::
        :context: close-figs

        >>> from numpy.random import choice
        >>> import matplotlib.pyplot as plt
        >>> from geneview import venn
        >>> musicians = {
        ... "Members of The Beatles": {"Paul McCartney", "John Lennon", "George Harrison", "Ringo Starr"},
        ... "Guitarists": {"John Lennon", "George Harrison", "Jimi Hendrix", "Eric Clapton", "Carlos Santana"},
        ... "Played at Woodstock": {"Jimi Hendrix", "Carlos Santana", "Keith Moon"}}
        >>> ax = venn(musicians)

   Rename the labels for each petal by manual and set the color of legend text to be the
   same as petal by setting bool variable as ``legend_use_petal_color=True``.

    .. plot::
        :context: close-figs

        >>> ax = venn(musicians, names=["A", "B", "C"], legend_use_petal_color=True)

    Examples of Venn diagrams for various numbers of sets.

    Venn diagrams can be plotted for 2, 3, 4, or 5 sets using ellipses, and for 6 sets using triangles.
    The venn() function accepts optional arguments data, names, fmt, palette, alpha, fontsize,
    legend_loc and ax.

    .. plot::
        :context: close-figs

        >>> from itertools import chain, islice
        >>> from string import ascii_uppercase
        >>> from numpy.random import choice
        >>> from geneview import venn
        >>> _, top_axs = plt.subplots(ncols=3, nrows=1, figsize=(18, 5))
        >>> _, bot_axs = plt.subplots(ncols=2, nrows=1, figsize=(18, 8))
        >>> cmaps = ["cool", list("rgb"), "plasma", "viridis", "Set1"]
        >>> letters = iter(ascii_uppercase)
        >>> for n_sets, cmap, ax in zip(range(2, 7), cmaps, chain(top_axs, bot_axs)):
        ...    dataset_dict = {name: set(choice(1000, 700, replace=False)) for name in islice(letters, n_sets)}
        ...    _ = venn(dataset_dict,
        ...             fmt="{percentage:.1f}%",  # "{size}", "{logic}"
        ...             palette=cmap,
        ...             fontsize=12,
        ...             legend_use_petal_color=True,
        ...             legend_loc="upper left",
        ...             ax=ax)

    If necessary, the labels on the petals (i.e., various intersections in the Venn diagram) can be adjusted manually.
    For this, generate_petal_labels() can be called first to get the petal_labels dictionary, or you can deal the data
    by yourself, just make sure that the data is dictionary and the format should be like:

        {'001': '0',   # The keys of dict is consist by 0 and 1, and the value should be str type.
         '010': '5',
         '011': '0',
         '100': '3',
         '101': '2',
         '110': '2',
         '111': '3'}

    After modification, pass petal_labels to functions venn().

    >>> from geneview import generate_petal_labels
    >>> dataset_dict = {name: set(choice(1000, 250, replace=False)) for name in list("ABCD")}
    >>> petal_labels = generate_petal_labels(dataset_dict.values(), fmt="{logic}\\n({percentage:.1f}%)")
    >>> ax = venn(data=petal_labels, names=list(dataset_dict.keys()), legend_use_petal_color=True)
    """
    if is_already_venn_dataset(data, names):
        return vennx(data=data,
                     names=names,
                     palette=palette,
                     alpha=alpha,
                     fontsize=fontsize,
                     legend_use_petal_color=legend_use_petal_color,
                     legend_loc=legend_loc, ax=ax)

    elif not is_valid_dataset_dict(data):
        raise TypeError("Only dictionaries of sets are understood")

    return vennx(data=generate_petal_labels(data.values(), fmt=fmt),
                 names=list(data.keys()) if names is None else names,
                 palette=palette,
                 alpha=alpha,
                 fontsize=fontsize,
                 legend_use_petal_color=legend_use_petal_color,
                 legend_loc=legend_loc,
                 ax=ax)



"""Plot admixture

Author: Shujia Huang
Date: 2021-05-01
"""
import warnings
from itertools import cycle
import numpy as np
import pandas as pd

from matplotlib.pyplot import subplots

from ..algorithm import hierarchical_cluster
from ..palette import generate_colors_palette


def _draw_admixtureplot(
        data=None,
        group_order=None,
        linewidth=1.0,
        palette="tab10",
        xticklabels=None,
        xticklabel_kws=None,
        ylabel=None,
        ylabel_kws=None,
        hierarchical_kws=None,
        ax=None
):
    """Plot admixture.

    Parameters
    ----------
    data : dict.
        Input data structure. The key of `data` is the name of sub population and the
        value of `data` is a pandas.DataFrame which contain the admixture output (.Q)
        of the specify population. e.g:
        data = {"sub-pop1": pandas.DataFrame("admixture result(.Q) for 'sub-pop1'"),
                "sub-pop2": pandas.DataFrame("admixture result(.Q) for 'sub-pop2'")}

    group_order : vector of strings, optional
        Specify the order of processing and plotting for the estimating sub populations.
    """
    if ax is None:
        _, ax = subplots(1, 1, figsize=(14, 2), facecolor="w", constrained_layout=True)

    if hierarchical_kws is None:
        hierarchical_kws = {"method": "average", "metric": "euclidean"}
    if "axis" not in hierarchical_kws:
        hierarchical_kws["axis"] = 0  # row axis (by sample) to use to calculate cluster.

    if group_order is None:
        group_order = list(set(data.keys()))

    n = 0
    for g in group_order:
        n += len(data[g])
        hc = hierarchical_cluster(data=data[g], **hierarchical_kws)
        data[g] = hc.data.iloc[hc.reordered_index]  # re-order the data.

    x = np.arange(n)
    base_y = np.zeros(len(x))

    column_names = data[group_order[0]].columns
    colors = generate_colors_palette(cmap=palette, n_colors=len(column_names))
    palette = cycle(colors)
    if len(colors) < len(column_names):
        msg = ("The categories of colors setting by `palette` is less than "
               "the number of estimating sub populations (K) in admixture, "
               "which could cause a confuse plot. Please reset the palette.")
        warnings.warn(msg)

    for k in column_names:
        c = next(palette)  # one color for one 'k'
        start_g_pos = 0
        add_y = []
        for g in group_order:  # keep group order
            try:
                y = data[g][k]
                end_g_pos = start_g_pos + len(y)
                ax.bar(x[start_g_pos:end_g_pos], y, bottom=base_y[start_g_pos:end_g_pos],
                       color=c, width=1.0, linewidth=0)

                start_g_pos = end_g_pos
                add_y.append(y)
            except KeyError:
                raise KeyError("KeyError: '%s'. Missing in 'group_order'." % g)

        base_y += np.array(pd.concat(add_y, axis=0))

    g_pos = 0
    xticks_pos = []
    for g in group_order:  # make group order
        g_size = len(data[g])
        xticks_pos.append(g_pos + 0.5 * g_size)

        g_pos += g_size
        ax.axvline(x=g_pos, color="k", lw=linewidth)

    ax.spines["left"].set_linewidth(linewidth)
    ax.spines["top"].set_linewidth(linewidth)
    ax.spines["right"].set_linewidth(linewidth)
    ax.spines["bottom"].set_linewidth(linewidth)

    ax.set_xlim([x[0], x[-1]])
    ax.set_ylim([0, 1.0])
    ax.tick_params(bottom=False, top=False, left=False, right=False)

    if xticklabel_kws is None:
        xticklabel_kws = {}

    if xticklabels and (len(group_order) != len(xticklabels)):
        raise ValueError("The size of 'xticklabels'(%d) is not the same with "
                         "'group_order'(%d)." % (len(xticklabels), len(group_order)))

    ax.set_xticks(xticks_pos)
    ax.set_xticklabels(group_order if xticklabels is None else xticklabels,
                       **xticklabel_kws)

    if ylabel_kws is None:
        ylabel_kws = {}

    ax.set_ylabel(ylabel, **ylabel_kws)
    ax.set_yticks([])

    return ax


def _load_admixture_from_file(in_admixture_fname, in_sample_info_fname, shuffle_popsample_kws=None):
    if ("axis" in shuffle_popsample_kws) and (shuffle_popsample_kws["axis"] == 1):
        warnings.warn("axis=1 means sampling the data by columns, Which is not "
                      "allow and may not be right in admixture data.")

    df = pd.read_table(in_admixture_fname, sep=" ", header=None)
    sample_info = pd.read_table(in_sample_info_fname, header=None, names=["Group"])
    popset = set(sample_info["Group"])

    if len(sample_info) != len(df):
        raise ValueError("The size of sample_info(%d) and input admixture result(%d) "
                         "must be the same." % (len(sample_info), len(df)))

    data = {}
    for g in popset:
        g_data = df[sample_info["Group"] == g].copy()
        g_size = len(g_data)
        if shuffle_popsample_kws:
            shuffle_raw_n = shuffle_popsample_kws["n"] if "n" in shuffle_popsample_kws else None
            if shuffle_raw_n and shuffle_raw_n > g_size:
                shuffle_popsample_kws["n"] = g_size
                data[g] = g_data.sample(**shuffle_popsample_kws)
                shuffle_popsample_kws["n"] = shuffle_raw_n  # reset to the raw N
            else:
                data[g] = g_data.sample(**shuffle_popsample_kws)
        else:
            data[g] = g_data

    return data


def admixtureplot(
        data,
        population_info=None,
        shuffle_popsample_kws=None,
        group_order=None,
        linewidth=1.0,
        palette="tab10",
        xticklabels=None,
        xticklabel_kws=None,
        ylabel=None,
        ylabel_kws=None,
        hierarchical_kws=None,
        ax=None
):
    """Plot admixture.

        Parameters
        ----------
        data : a file path to the admixture output(.Q) or a dict.
            Input data structure.
            The key of `data` is the name of sub population and the value of `data`
            is a pandas.DataFrame which contain the admixture output (.Q) of the
            specify population if `data` is a dict. e.g:
            data = {"sub-pop1": pandas.DataFrame("admixture result(.Q) for 'sub-pop1'"),
                    "sub-pop2": pandas.DataFrame("admixture result(.Q) for 'sub-pop2'")}

        population_info : String or None, optional
            File path to a sample information list, each line only contain a group information
            for admixture row.

        shuffle_popsample_kws : key, value pairings, or None, optional
            The keyword argument of `pandas.sample` method to sample rows (in random order) for
            each group of population. If None, keep the raw data.

        group_order : vector of strings, optional
            Specify the order of processing and plotting for the estimating sub populations.

        linewidth : float, optional, default: 1.0
            Set the line width in plot

        palette : string, list, or :class:`matplotlib.colors.Colormap`, optional, default: tab10.
            String values are passed to :func:`generate_colors`. List values imply categorical
            mapping, while a colormap object implies numeric mapping.

        xticklabels : list, optional
            The label texts of x-axis.

        xticklabel_kws : key, value pairings, or None, optional
            Other keyword arguments are passed to set xtick labels in
            maplotlib.axis.Axes.set_xticklabels.

        ylabel : string, optional
            Set the y axis label of the current axis. The label will set to be the
            column number of admixture output if ylabel is None.

        ylabel_kws : key, value pairings, or None, optional
            Other keyword arguments are passed to set y label in
            ``maplotlib.axis.Axes.set_ylabel``.

        hierarchical_kws : key, value pairings, or None, optional
            Other keyword arguments are passed to set the hierarchical clustering in
            ``geneview.algorithm.hierarchical_cluster``.

        ax : matplotlib axis, optional
            Axis to plot on, otherwise define a subplot by ``admixtureplot()``.

        Returns
        -------
        ax : matplotlib Axes
            Axes object with the admixtureplot.

        Examples
        --------

        A simple admixture plot example.

        .. plot::
            :context: close-figs

            >>> from geneview.utils import load_dataset
            >>> from geneview import admixtureplot
            >>> admixture_fn = load_dataset("admixture_output.Q")
            >>> population_fn = load_dataset("admixture_population.info")
            >>> ax = admixtureplot(data=admixture_fn, population_info=population_fn)

        Plot the admixture by a specify population order

        .. plot::
            :context: close-figs

            >>> pop_group_1kg = ["KHV", "CDX", "CHS", "CHB", "JPT", "BEB", "STU", "ITU",
            ...                  "GIH", "PJL", "FIN", "CEU", "GBR", "IBS", "TSI", "PEL",
            ...                  "PUR", "MXL", "CLM", "ASW", "ACB", "GWD", "MSL", "YRI",
            ...                  "ESN", "LWK"]
            >>> ax = admixtureplot(data=admixture_fn, population_info=population_fn,
            ...                    group_order=pop_group_1kg)

        Setting the ``shuffle_popsample_kws`` argument if you wish to sample some samples for each
        population group in the admixture plot.

        only sampling 80 samples for each population group.
        .. plot::
            :context: close-figs

            >>> ax = admixtureplot(data=admixture_fn,
            ...                    population_info=population_fn,
            ...                    shuffle_popsample_kws={"n": 80},
            ...                    group_order=pop_group_1kg)

        Or specifies the fraction of each group population.

        .. plot::
            :context: close-figs

            >>> ax = admixtureplot(data=admixture_fn,
            ...                    population_info=population_fn,
            ...                    shuffle_popsample_kws={"frac": 0.2},
            ...                    group_order=pop_group_1kg)

        0.2 in `shuffle_popsample_kws` means shuffle 20% of the data for each estimate poplation group
        in random order.

        ``admixtureplot()`` also allow you to define your own data (in dict type) by manual.

        .. plot::
            :context: close-figs

            >>> df = pd.read_table(admixture_fn, sep=" ", header=None)
            >>> sample_info = pd.read_table(population_fn, sep="\\t", header=None, names=["Group"])
            >>> pop_set = set(sample_info["Group"])
            >>> data = {}
            >>> for g in pop_set:
            ...     g_data = df[sample_info["Group"]==g].copy()
            ...     # Sub sampling: keep less than 140 samples for each group
            ...     data[g] = g_data.sample(n=140, random_state=100) if len(g_data)>140 else g_data

            # Plot the figure
            >>> import matplotlib.pyplot as plt
            >>> f, ax = plt.subplots(1, 1, figsize=(14, 3), facecolor="w", constrained_layout=True, dpi=300)
            >>> ax = admixtureplot(data=data,
            ...                    group_order=pop_group_1kg,
            ...                    palette="Set1",
            ...                    xticklabel_kws={"rotation": "vertical"},
            ...                    ylabel="K=11",
            ...                    ylabel_kws={"rotation": 0, "ha": "right"},
            ...                    ax=ax)

    """
    if shuffle_popsample_kws is None:
        shuffle_popsample_kws = {}

    if (population_info is not None) and (not isinstance(population_info, str)):
        raise ValueError("`population_info` must be a file path which contain "
                         "the population information for each row of admixture "
                         "result. One element per line.")

    if (population_info is None) and (not isinstance(data, dict)):
        raise ValueError("``data`` must be a dictionary and the key should be the group "
                         "information, values should be a dataframe which contain the result "
                         "of admixture if the ``population_info`` argument is None.")

    if (not isinstance(data, dict)) and (not isinstance(data, str)):
        raise ValueError("`data` should be a dict or a file path to the admixture output(.Q).")

    if isinstance(data, str):
        data = _load_admixture_from_file(
            data, population_info, shuffle_popsample_kws=shuffle_popsample_kws
        )

    # infer ylabel if ylabel is None
    if ylabel is None:
        g = list(data.keys())[0]
        ylabel = "K=%d" % len(data[g].columns)

    return _draw_admixtureplot(data=data,
                               group_order=group_order,
                               linewidth=linewidth,
                               palette=palette,
                               xticklabels=xticklabels,
                               xticklabel_kws=xticklabel_kws,
                               ylabel=ylabel,
                               ylabel_kws=ylabel_kws,
                               hierarchical_kws=hierarchical_kws,
                               ax=ax)



"""
Utility functionality (:mod:`geneview.util`)
============================================

This package provides general exception/warning definitions used throughout
geneview, as well as various utility functionality, including plotting and
unit-testing convenience functions.

"""
from ._misc import is_numeric, is_integer, iqr, _kde_support, \
    categorical_order, freedman_diaconis_bins
from ._plott import General
from ._palette import desaturate, set_hls_values, get_color_cycle
from ._dataset import get_dataset_names, load_dataset
from ._adjust_text import adjust_text
from ._decorators import deprecate_positional_args

__all__ = ["is_numeric",
           "is_integer",
           "iqr",
           "_kde_support",
           "General",
           "desaturate",
           "get_dataset_names",
           "load_dataset",
           "set_hls_values",
           "categorical_order",
           "get_color_cycle",
           "freedman_diaconis_bins",
           "adjust_text",
           "deprecate_positional_args"]



"""
Utility functions for getting datasets from geneview online dataset repository.
"""
import os
import re
import pandas as pd
from urllib.request import urlopen, urlretrieve

from ._misc import is_numeric


def get_dataset_names():
    """Report available example datasets, useful for reporting issues."""

    url = "https://github.com/ShujiaHuang/geneview-data"
    with urlopen(url) as resp:
        html = resp.read()

    pat = r"/ShujiaHuang/geneview-data/blob/master/(\w*).csv"
    datasets = re.findall(pat, html.decode())
    return datasets


def load_dataset(name, cache=True, data_home=None, **kws):
    """Load a dataset from the online repository (requires internet).

    Parameters
    ----------
    name : str
        Name of the dataset (`name`.csv on
        https://github.com/ShujiaHuang/geneview-data).  You can obtain list of
        available datasets using :func:`get_dataset_names`

    cache : boolean, optional
        If True, then cache data locally and use the cache on subsequent calls

    data_home : string, optional
        The directory in which to cache data. By default, uses ~/geneview_data/

    kws : dict, optional
        Passed to pandas.read_csv

    Examples
    --------
    Load the preview data of GWAS

        >>> import geneview as gv
        >>> gwas_data = gv.utils.load_dataset("gwas")
        >>> file_path = gv.utils.load_dataset("bcftools.vcf.stats")  # only return a path name

    """
    path_csv = "https://raw.githubusercontent.com/ShujiaHuang/geneview-data/master/{0}.csv"
    path_full = "https://raw.githubusercontent.com/ShujiaHuang/geneview-data/master/{0}"
    path_name = path_full.format(name) if "." in name else path_csv.format(name)

    if cache:
        cache_path = os.path.join(_get_data_home(data_home), os.path.basename(path_name))
        if not os.path.exists(cache_path):
            urlretrieve(path_name, cache_path)
        path_name = cache_path

    if path_name.endswith(".csv"):
        df = pd.read_csv(path_name, **kws)
        if df.iloc[-1].isnull().all():
            df = df.iloc[:-1]
        return df
    else:
        return path_name


def _tr(s):
    """ transform numeric to be float data. """
    return float(s) if is_numeric(s) else s


def _get_data_home(data_home=None):
    """Return the path of the geneview data directory.

    This is used by the ``load_dataset`` function.

    If the ``data_home`` argument is not specified, the default location
    is ``~/geneview-data``.

    Alternatively, a different default location can be specified using the
    environment variable ``GENEVIEW_DATA``.
    """
    if data_home is None:
        data_home = os.environ.get('GENEVIEW_DATA', os.path.join('~', 'geneview-data'))
    data_home = os.path.expanduser(data_home)
    if not os.path.exists(data_home):
        os.makedirs(data_home)

    return data_home



from .palette import *
from .ext import *

from .karyotype import karyoplot
from .baseplot import venn, generate_petal_labels
from .gwas import manhattanplot, qqplot, qqnorm
from .popgene import admixtureplot



import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt


def palplot(pal, size=1):
    """Plot the values in a color palette as a horizontal array.

    Parameters
    ----------
    pal : sequence of matplotlib colors
        colors
    size :
        scaling factor for size of plot

    """
    n = len(pal)
    f, ax = plt.subplots(1, 1, figsize=(n * size, size))
    ax.imshow(np.arange(n).reshape(1, n),
              cmap=mpl.colors.ListedColormap(list(pal)),
              interpolation="nearest", aspect="auto")
    ax.set_xticks(np.arange(n) - .5)
    ax.set_yticks([-.5, .5])
    ax.set_xticklabels([])
    ax.set_yticklabels([])


def puppyplot(grown_up=False):
    """Plot today's daily puppy. Only works in the IPython notebook."""
    from urllib.request import urlopen
    from IPython.display import HTML

    try:
        from bs4 import BeautifulSoup
        url = "http://www.dailypuppy.com"
        if grown_up:
            url += "/dogs"
        html_doc = urlopen(url)
        soup = BeautifulSoup(html_doc, 'lxml')
        puppy = soup.find("div", {"class": "daily_puppy"})
        return HTML(str(puppy.img))
    except ImportError:
        html = ('<img src="http://cdn-www.dailypuppy.com/media/dogs/'
                '77165_20151221_023650412_iOS.jpg_w750.jpg" '
                'style="width:450px;"/>')
        return HTML(html)



from matplotlib.colors import to_rgba
from matplotlib.cm import ScalarMappable


def generate_colors_palette(cmap="viridis", n_colors=10, alpha=1.0):
    """Generate colors from matplotlib colormap; pass list to use exact colors"""
    if isinstance(cmap, list):
        colors = [list(to_rgba(color, alpha=alpha)) for color in cmap]
    else:
        scalar_mappable = ScalarMappable(cmap=cmap)
        colors = scalar_mappable.to_rgba(range(n_colors), alpha=alpha).tolist()
    return colors



# Copyright (C) 2016-2021 Shujia Huang <huangshujia9@gmail.com>
# temporarily redirect config directory to prevent matplotlib importing
# testing that for writeable directory which results in sandbox error in
# certain easy_install versions
import os
from argparse import Namespace


DESCRIPTION = "Geneview: A python package for genomics data visualization."
meta = Namespace(
    __DISTNAME__     = "geneview",
    __AUTHOR__       = "Shujia Huang",
    __AUTHOR_EMAIL__ = "huangshujia9@gmail.com",
    __URL__          = "https://github.com/ShujiaHuang/geneview",
    __LICENSE__      = "BSD (3-clause)",
    __DOWNLOAD_URL__ = "https://github.com/ShujiaHuang/geneview",
    __VERSION__      = "0.1.0",
)

try:
    from setuptools import setup, find_packages
    _has_setuptools = True
except ImportError:
    from distutils.core import setup, find_packages


if __name__ == "__main__":

    THIS_PATH = os.path.abspath(os.path.dirname(__file__))
    long_description = os.path.join(THIS_PATH, "README.md")
    #with open(os.path.join(THIS_PATH, "README.md"), encoding="utf-8") as f:
    #    LONG_DESCRIPTION = f.read()

    setup(name=meta.__DISTNAME__,
          author=meta.__AUTHOR__,
          author_email=meta.__AUTHOR_EMAIL__,
          maintainer=meta.__AUTHOR__,
          maintainer_email=meta.__AUTHOR_EMAIL__,
          description=DESCRIPTION,
          long_description=(open(long_description).read()),
          long_description_content_type="text/markdown",
          license=meta.__LICENSE__,
          url=meta.__URL__,
          version=meta.__VERSION__,
          download_url=meta.__DOWNLOAD_URL__,
          install_requires=[
              "numpy",
              "scipy",
              "pandas",
              "matplotlib",
              "seaborn",
          ],
          packages=find_packages(),
          classifiers=[
             "Intended Audience :: Science/Research",
             "Programming Language :: Python :: 3.7",
             "Programming Language :: Python :: 3.8",
             "License :: OSI Approved :: BSD License",
             "Topic :: Scientific/Engineering :: Bio-Informatics",
             "Topic :: Scientific/Engineering :: Visualization",
             "Topic :: Multimedia :: Graphics",
             "Operating System :: POSIX",
             "Operating System :: Unix",
             "Operating System :: MacOS"],
          )



"""
Read more about seaborn color palette
http://stanford.edu/~mwaskom/software/seaborn-dev/tutorial/color_palettes.html
"""
from ._xkcd_rgb import xkcd_rgb
from ._circos import circos

from ._palettes import generate_colors_palette




# -*- coding: utf-8 -*-
from math import log
import numpy as np
from matplotlib import cm
import matplotlib.pyplot as plt


def _iter_point(c, iter_num=100, escape_radius=2):
    z = c
    n = 1
    for i in range(1, iter_num):  # 最多迭代100次
        n = i
        if abs(z) > escape_radius:
            break  # 半径大于该值则认为逃逸
        z = z * z + c

    return n  # 迭代次数


def _smooth_iter_point(c, iter_num=20, escape_radius=10):
    z = c
    n = 1
    for i in range(1, iter_num):  # 最多迭代100次
        n = i
        if abs(z) > escape_radius:
            break  # 半径大于该值则认为逃逸
        z = z * z + c

    mu = n - log(log(abs(z), 2), 2) if abs(z) > 2.0 else n
    return mu  # 迭代次数


def _mandelbrot(cx, cy, d, n=200, iter_point_func=_iter_point,
                iter_num=100, escape_radius=2):
    """                                                                                  
    绘制点 (cx, cy) 附近正负d的范围的 Mandelbrot
    """
    x0, x1, y0, y1 = cx - d, cx + d, cy - d, cy + d
    y, x = np.ogrid[y0:y1:n * 1j, x0:x1:n * 1j]
    c = x + y * 1j

    m = np.frompyfunc(iter_point_func, 3, 1)(c, iter_num, escape_radius).astype(np.float)
    return m


def mandelbrot():
    plt.subplot(231)
    plt.imshow(_mandelbrot(-0.5, 0, 1.5),
               cmap=cm.Blues_r,
               extent=[-0.5 - 1.5, -0.5 + 1.5, -1.5, 1.5])
    plt.gca().set_axis_off()

    x, y = 0.27322626, 0.595153338
    for i in range(2, 7):
        plt.subplot(230 + i)
        _mandelbrot(x, y, 0.2 ** (i - 1))
        d = 0.2 ** (i - 1)
        plt.imshow(_mandelbrot(x, y, d),
                   cmap=cm.Blues_r, extent=[x - d, x + d, y - d, y + d])
        plt.gca().set_axis_off()

    plt.subplots_adjust(0.02, 0, 0.98, 1, 0.02, 0)


def mandelbrot_smooth():
    x, y, d = -0.5, 0, 1.5
    escape_radius, iter_num, N = 10, 20, 300

    plt.subplot(121)
    plt.imshow(_mandelbrot(x, y, d, N, _iter_point,
                           iter_num, escape_radius),
               cmap=cm.Blues_r, extent=[x - d, x + d, y - d, y + d])
    plt.gca().set_axis_off()

    plt.subplot(122)
    plt.imshow(_mandelbrot(x, y, d, N, _smooth_iter_point,
                           iter_num, escape_radius),
               cmap=cm.Blues_r, extent=[x - d, x + d, y - d, y + d])

    plt.subplots_adjust(0.02, 0, 0.98, 1, 0.02, 0)
    plt.gca().set_axis_off()


if __name__ == "__main__":
    # mandelbrot()
    mandelbrot_smooth()
    plt.show()
